# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

"""Initial contents of the SciCat testing backend.

The contents of this module are used to initialize
the local database before making it available to tests.
"""

import pickle
from copy import deepcopy
from pathlib import Path
from typing import Dict, List, Union

from dateutil.parser import parse as parse_datetime

from ... import model
from ...client import Client
from ...filesystem import RemotePath
from ...model import (
    DatasetType,
    DownloadAttachment,
    DownloadDataset,
    DownloadOrigDatablock,
    UploadAttachment,
    UploadDataFile,
    UploadDerivedDataset,
    UploadOrigDatablock,
    UploadRawDataset,
    UploadTechnique,
)
from ...pid import PID
from ...thumbnail import Thumbnail
from .config import SITE, SciCatAccess, SciCatUser

# Dataset models to upload to the database.
_DATASETS: Dict[str, Union[UploadRawDataset, UploadDerivedDataset]] = {
    "raw": UploadRawDataset(
        ownerGroup="PLACEHOLDER",
        accessGroups=["uu", "faculty"],
        classification="IN=medium,AV=low,CO=low",
        contactEmail="ponder.stibbons@uu.am",
        creationTime=parse_datetime("2004-06-13T01:45:28.100Z"),
        datasetName="My darkest magic yet",
        description="Doing some dark shit",
        isPublished=False,
        numberOfFiles=2,
        numberOfFilesArchived=0,
        owner="PLACEHOLDER",
        ownerEmail="PLACE@HOLD.ER",
        size=619,
        sourceFolder=RemotePath("/hex/data/123"),
        type=DatasetType.RAW,
        principalInvestigator="Ponder Stibbons",
        creationLocation=SITE,
        techniques=[UploadTechnique(pid="DM666", name="dark_magic")],
        scientificMetadata={
            "data_type": "event data",
            "temperature": {"value": "123", "unit": "K"},
            "weight": {"value": "42", "unit": "mg"},
        },
    ),
    "derived": UploadDerivedDataset(
        ownerGroup="PLACEHOLDER",
        accessGroups=["uu", "faculty"],
        classification="IN=medium,AV=low,CO=low",
        contactEmail="ponder.stibbons@uu.am",
        creationTime=parse_datetime("2005-11-04T13:37:44.002Z"),
        datasetName="Reprocessed dark magic",
        description="Making it even darker",
        isPublished=True,
        numberOfFiles=1,
        numberOfFilesArchived=0,
        owner="PLACEHOLDER",
        ownerEmail="PLACE@HOLD.ER",
        size=464,
        sourceFolder=RemotePath("/hex/data/dd"),
        type=DatasetType.DERIVED,
        investigator="Ponder Stibbons",
        inputDatasets=[],
        usedSoftware=["scitacean"],
        scientificMetadata={
            "data_type": "reduced",
            "pressure": {"value": "8.1", "unit": "Pa"},
        },
    ),
    "public": UploadRawDataset(
        ownerGroup="PLACEHOLDER",
        accessGroups=["uu"],
        classification="IN=medium,AV=low,CO=low",
        contactEmail="mustrum.ridcully69@uu.am",
        creationTime=parse_datetime("1998-11-05T23:00:42.000Z"),
        datasetName="Shoe counter",
        description="Got all these shoes!",
        isPublished=True,
        numberOfFiles=1,
        numberOfFilesArchived=0,
        owner="PLACEHOLDER",
        ownerEmail="PLACE@HOLD.ER",
        size=64,
        sourceFolder=RemotePath("/hex/secret/stuff"),
        type=DatasetType.RAW,
        principalInvestigator="Mustrum Ridcully",
        creationLocation=SITE,
        techniques=[UploadTechnique(pid="S", name="shoes")],
    ),
    "partially-broken": model.construct(
        UploadDerivedDataset,
        _strict_validation=False,
        _quiet=True,
        ownerGroup="PLACEHOLDER",
        accessGroups=["uu"],
        classification="IN=medium,AV=low,CO=low",
        contactEmail="owner@mail.com",
        orcidOfOwner="00-11-22-33",
        creationTime=parse_datetime("1998-11-05T23:00:42.000Z"),
        datasetName="Dataset with broken fields",
        description="Bad fields: orcidOfOwner, numberOfFiles, size",
        isPublished=False,
        numberOfFiles=-1,
        numberOfFilesArchived=0,
        owner="PLACEHOLDER",
        ownerEmail="PLACE@HOLD.ER",
        size=-10,
        sourceFolder=RemotePath("/remote/source"),
        type=DatasetType.DERIVED,
        investigator="who?!",
        inputDatasets=[],
        usedSoftware=["scitacean"],
    ),
}

# Orig datablocks to upload to the database.
_ORIG_DATABLOCKS: Dict[str, List[UploadOrigDatablock]] = {
    "raw": [
        UploadOrigDatablock(
            datasetId=PID(pid="PLACEHOLDER"),
            ownerGroup="PLACEHOLDER",
            size=619,
            accessGroups=["uu", "faculty"],
            chkAlg="md5",
            dataFileList=[
                UploadDataFile(
                    path="file1.txt",
                    size=300,
                    time=parse_datetime("2005-11-04T13:22:09.000Z"),
                    chk="97157d347fe9af920f5e61e96cf401cb",
                    gid="1000",
                    uid="1000",
                    perm="777",
                ),
                UploadDataFile(
                    path="sub/song.mp3",
                    size=319,
                    time=parse_datetime("2005-11-02T09:02:54.100Z"),
                    chk=None,
                    gid="1000",
                    uid="1000",
                    perm="777",
                ),
            ],
        )
    ],
    "derived": [
        UploadOrigDatablock(
            datasetId=PID(pid="PLACEHOLDER"),
            ownerGroup="PLACEHOLDER",
            size=464,
            accessGroups=["uu", "faculty"],
            chkAlg="sha256",
            dataFileList=[
                UploadDataFile(
                    path="table.csv",
                    size=464,
                    time=parse_datetime("2005-10-31T00:00:01.000Z"),
                    chk="dddd8355da9105acabb9928196f022ca0581ffb73d8b89c891eb6f71477cb4cb",
                    gid="2000",
                    uid="0",
                    perm="656",
                ),
            ],
        )
    ],
    "public": [
        UploadOrigDatablock(
            datasetId=PID(pid="PLACEHOLDER"),
            ownerGroup="PLACEHOLDER",
            size=64,
            accessGroups=["uu"],
            chkAlg="md5",
            dataFileList=[
                UploadDataFile(
                    path="shoes",
                    size=64,
                    time=parse_datetime("1998-11-05T22:56:13.000Z"),
                    chk="95fe96bf90f6a53c1e20d6578e0d9e6e",
                    gid="0",
                    uid="0",
                    perm="0",
                ),
            ],
        )
    ],
}

_ATTACHMENTS: Dict[str, List[UploadAttachment]] = {
    "derived": [
        UploadAttachment(
            caption="Process Overview",
            ownerGroup="PLACEHOLDER",
            accessGroups=["uu"],
            datasetId=PID(pid="PLACEHOLDER"),
            thumbnail=Thumbnail(mime="image/png", data=b"nag;aso;i"),
        ),
        UploadAttachment(
            caption="Pretty picture no 1",
            ownerGroup="PLACEHOLDER",
            accessGroups=["uu", "faculty"],
            datasetId=PID(pid="PLACEHOLDER"),
            thumbnail=Thumbnail(mime="image/jpeg", data=b"gj0ajs93jka2jv89a"),
            sampleId="kjsdf",
        ),
    ]
}

INITIAL_DATASETS: Dict[str, DownloadDataset] = {}
"""Initial datasets in the testing database."""
INITIAL_ORIG_DATABLOCKS: Dict[str, List[DownloadOrigDatablock]] = {}
"""Initial orig datablocks in the testing database."""
INITIAL_ATTACHMENTS: Dict[str, List[DownloadAttachment]] = {}


def _apply_config_dataset(
    dset: Union[UploadRawDataset, UploadDerivedDataset], user: SciCatUser
) -> Union[UploadRawDataset, UploadDerivedDataset]:
    dset = deepcopy(dset)
    dset.owner = user.username
    dset.ownerGroup = user.group
    dset.ownerEmail = user.email
    return dset


def _apply_config_orig_datablock(
    dblock: UploadOrigDatablock, dset: DownloadDataset, user: SciCatUser
) -> UploadOrigDatablock:
    dblock = deepcopy(dblock)
    dblock.ownerGroup = user.group
    dblock.datasetId = dset.pid  # type: ignore[assignment]
    return dblock


def _apply_config_attachment(
    attachment: UploadAttachment, user: SciCatUser
) -> UploadAttachment:
    attachment = deepcopy(attachment)
    attachment.ownerGroup = user.group
    return attachment


def _create_dataset_model(
    client: Client, dset: Union[UploadRawDataset, UploadDerivedDataset]
) -> DownloadDataset:
    uploaded = client.scicat.create_dataset_model(dset)
    # pid is a str if validation fails but we need a PID for fake clients.
    uploaded.pid = PID.parse(uploaded.pid)  # type: ignore[arg-type]
    return uploaded


def seed_database(*, client: Client, scicat_access: SciCatAccess) -> None:
    """Seed the database for testing.

    Uses the provided client to upload the datasets.
    Initializes ``INITIAL_DATASETS``, ``INITIAL_ORIG_DATABLOCKS``,
    and ``INITIAL_ATTACHMENTS`` with finalized datasets returned by the client.
    """
    upload_datasets = {
        key: _apply_config_dataset(dset, scicat_access.user)
        for key, dset in _DATASETS.items()
    }
    download_datasets = {
        key: _create_dataset_model(client, dset)
        for key, dset in upload_datasets.items()
    }
    INITIAL_DATASETS.update(download_datasets)

    upload_orig_datablocks = {
        key: [
            _apply_config_orig_datablock(
                dblock, download_datasets[key], scicat_access.user
            )
            for dblock in dblocks
        ]
        for key, dblocks in _ORIG_DATABLOCKS.items()
    }
    download_orig_datablocks = {
        key: [client.scicat.create_orig_datablock(dblock) for dblock in dblocks]
        for key, dblocks in upload_orig_datablocks.items()
    }
    INITIAL_ORIG_DATABLOCKS.update(download_orig_datablocks)

    upload_attachments = {
        key: [
            _apply_config_attachment(attachment, scicat_access.user)
            for attachment in attachments
        ]
        for key, attachments in _ATTACHMENTS.items()
    }
    download_attachments = {
        key: [
            client.scicat.create_attachment_for_dataset(
                attachment,
                dataset_id=download_datasets[key].pid,  # type: ignore[arg-type]
            )
            for attachment in attachments
        ]
        for key, attachments in upload_attachments.items()
    }
    INITIAL_ATTACHMENTS.update(download_attachments)


def save_seed(target_dir: Path) -> None:
    """Save the processed seed to a file."""
    with open(target_dir / "seed", "wb") as f:
        pickle.dump(
            {
                "datasets": INITIAL_DATASETS,
                "orig_datablocks": INITIAL_ORIG_DATABLOCKS,
                "attachments": INITIAL_ATTACHMENTS,
            },
            f,
        )


def seed_worker(target_dir: Path) -> None:
    """Load the processed seed from a file."""
    with open(target_dir / "seed", "rb") as f:
        loaded = pickle.load(f)  # noqa: S301
        INITIAL_DATASETS.update(loaded["datasets"])
        INITIAL_ORIG_DATABLOCKS.update(loaded["orig_datablocks"])
        INITIAL_ATTACHMENTS.update(loaded["attachments"])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

"""Generate Python classes for SciCat models."""
import argparse
import subprocess
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Dict, Generator

from jinja2 import Environment, FileSystemLoader, Template
from spec import DatasetSpec, Spec, load_specs
from templates import BANNER

# Set to the URL of the JSON schema.
# See the README for details.
SCHEMA_URL = "http://localhost:3000/explorer-json"

# The root directory of the Scitacean repository.
# Modify this if the current file has moved.
REPO_ROOT = Path(__file__).resolve().parent.parent.parent
# Output file for model.py
MODEL_OUT_PATH = REPO_ROOT / "src" / "scitacean" / "model.py"
# Output file for _dataset_fields.py
DSET_FIELDS_OUT_PATH = REPO_ROOT / "src" / "scitacean" / "_dataset_fields.py"


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="generate_models.py", description="Generate models for Scitacean"
    )
    parser.add_argument(
        "--launch-scicat",
        action="store_true",
        help="Launch a temporary SciCat backend with docker",
    )
    return parser.parse_args()


def quote(s: str) -> str:
    if '"' in s:
        return f"'{s}'"
    return f'"{s}"'


def _template(name: str) -> Template:
    environment = Environment(  # noqa: S701
        loader=FileSystemLoader(Path(__file__).resolve().parent / "templates/"),
    )
    environment.filters["quote"] = quote
    return environment.get_template(f"{name}.py.jinja")


def _model_template() -> Template:
    return _template("model")


def _dataset_fields_template() -> Template:
    return _template("dataset_fields")


def generate_models(specs: Dict[str, Spec]) -> str:
    specs = dict(specs)
    dset_spec = specs.pop("Dataset")
    return _model_template().render(banner=BANNER, specs=specs, dset_spec=dset_spec)


def generate_dataset_fields(dset_spec: DatasetSpec) -> str:
    return _dataset_fields_template().render(banner=BANNER, spec=dset_spec)


def format_with_black(path: Path) -> None:
    # Root of Scitacean repo
    base_path = Path(__file__).resolve().parent.parent.parent
    subprocess.check_call(
        ["black", path.resolve().relative_to(base_path)],  # noqa: S603, S607
        cwd=base_path,
    )


@contextmanager
def _scicat_backend() -> Generator[None, None, None]:
    from scitacean.testing import backend

    with tempfile.TemporaryDirectory() as work_dir:
        docker_file = Path(work_dir) / "docker-compose.yaml"
        backend.configure(docker_file)
        backend.start_backend(docker_file)
        try:
            backend.wait_until_backend_is_live(max_time=20, n_tries=20)
            yield
        finally:
            backend.stop_backend(docker_file)


def load(real_backend: bool) -> Dict[str, Any]:
    if not real_backend:
        return load_specs(SCHEMA_URL)

    with _scicat_backend():
        from scitacean.testing.backend import config

        return load_specs(f"http://localhost:{config.SCICAT_PORT}/explorer-json")


def main() -> None:
    args = _parse_args()
    specs = load(args.launch_scicat)
    dset_spec = specs["Dataset"]

    with open(MODEL_OUT_PATH, "w") as f:
        f.write(generate_models(specs))
    format_with_black(MODEL_OUT_PATH)

    with open(DSET_FIELDS_OUT_PATH, "w") as f:
        f.write(generate_dataset_fields(dset_spec))
    format_with_black(DSET_FIELDS_OUT_PATH)


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
# ruff: noqa: S603, S607
"""Helpers to manage docker containers.

Primarily meant for testing.
"""
import json
import os
import subprocess
from typing import Any, Union

_PathLike = Union[str, os.PathLike]


def docker_compose_up(config_file: _PathLike, *services: str) -> None:
    subprocess.check_call(
        [
            "docker",
            "compose",
            "--file",
            os.fspath(config_file),
            "up",
            "--detach",
            "--force-recreate",
            *services,
        ]
    )


def docker_compose_down(config_file: _PathLike) -> None:
    subprocess.check_call(
        ["docker", "compose", "--file", os.fspath(config_file), "down", "--volumes"]
    )


def docker_compose_run(config_file: _PathLike, service: str, *cmd: str) -> None:
    subprocess.check_call(
        ["docker", "compose", "--file", os.fspath(config_file), "run", service, *cmd]
    )


def _try_parse_json(s: str) -> Any:
    # The precise output of `docker ps` can vary.
    # So parse it leniently; if a line isn't valid JSON, it surely does not
    # contain a container name.
    try:
        return json.loads(s)
    except json.JSONDecodeError:
        return None


def container_is_running(name: str) -> bool:
    ps = subprocess.run(
        ["docker", "ps", "--format=json"], stdout=subprocess.PIPE, check=True
    )
    lines = ps.stdout.decode(encoding="utf-8").split("\n")
    containers = [x for line in lines if (x := _try_parse_json(line))]
    for container in containers:
        names = container["Names"]
        if (isinstance(names, str) and names == name) or name in names:
            return True
    return False



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Hypothesis strategies for generating datasets.

This module defines a number of
`Hypothesis <https://hypothesis.readthedocs.io/en/latest/>`_ strategies to
generate test inputs.

Note
----
The ``datasets`` strategy can require a lot of time and memory due to the number of
parameters of ``Dataset``.
This can trip up Hypothesis' health checks, so you may need to disable them.
This can be done, e.g., by placing the following in ``conftest.py``:

.. code-block:: python

    import hypothesis

    hypothesis.settings.register_profile(
        "scitacean",
        suppress_health_check=[
            hypothesis.HealthCheck.data_too_large,
            hypothesis.HealthCheck.too_slow,
        ],
    )

Select the 'scitacean' profile during tests using the
``--hypothesis-profile=scitacean`` command line option.
"""

import string
from functools import partial
from typing import Any, Dict, Optional

from email_validator import EmailNotValidError, ValidatedEmail, validate_email
from hypothesis import strategies as st

from .. import PID, Dataset, DatasetType, RemotePath, model
from .._internal.orcid import orcid_checksum


# email_validator and by extension pydantic is more picky than hypothesis
# so make sure that generated emails actually pass model validation.
def _validate_email(email: str) -> Optional[ValidatedEmail]:
    try:
        return validate_email(email, check_deliverability=False)
    except EmailNotValidError:
        return None


def _is_valid_email(validated_email: Optional[ValidatedEmail]) -> bool:
    return validated_email is not None


def emails() -> st.SearchStrategy[str]:
    """A strategy for generating email addresses as strings.

    This differs from :func:`hypothesis.strategies.emails` in that it

    - all letters are lowercase,
    - uses only normalized email addresses, i.e., with utf-8 characters
      for internationalized names and never ASCII-encoding (see Punycode).

    This is done to match the behavior of ``pydantic.EmailStr`` and to allow for
    roundtrip tests through :class:`scitacean.Client`.

    Returns
    -------
    :
        String of email addresses.

    See Also
    --------
    scitacean.testing.strategies.multi_emails
    """
    return (
        st.emails()
        .map(lambda s: s.lower())
        .map(_validate_email)
        .filter(_is_valid_email)
        .map(lambda m: m.normalized)  # type: ignore[union-attr]
    )


def multi_emails(max_emails: int = 2) -> st.SearchStrategy[str]:
    """A strategy for generating multiple email addresses as strings.

    Parameters
    ----------
    max_emails:
        Maximum number of emails to generate.

    Returns
    -------
    :
        String of one or more, semicolon-separated email addresses.

    See Also
    --------
    scitacean.testing.strategies.emails
    """
    return st.lists(
        emails(),
        min_size=1,
        max_size=max_emails,
    ).map(lambda email: ";".join(email))


def _email_field_strategy(field: Dataset.Field) -> st.SearchStrategy[Optional[str]]:
    if field.required:
        return multi_emails()
    return st.none() | multi_emails()


def orcids() -> st.SearchStrategy[str]:
    """A strategy for generating ORCID ids.

    The generated ids are structurally valid, but do not necessarily exist.

    Returns
    -------
    :
        Strings of ORCID ids.
    """

    def make_orcid(digits: str) -> str:
        digits = digits[:-1] + orcid_checksum(digits)
        return "https://orcid.org/" + "-".join(
            digits[i : i + 4] for i in range(0, 16, 4)
        )

    return st.text(alphabet="0123456789", min_size=16, max_size=16).map(make_orcid)


def _orcid_field_strategy(field: Dataset.Field) -> st.SearchStrategy[Optional[str]]:
    if field.required:
        return orcids()
    return st.none() | orcids()


def _scientific_metadata_strategy(
    field: Dataset.Field,
) -> st.SearchStrategy[Dict[str, Any]]:
    return st.dictionaries(
        keys=st.text(),
        values=st.text() | st.dictionaries(keys=st.text(), values=st.text()),
    )


def _job_parameters_strategy(
    field: Dataset.Field,
) -> st.SearchStrategy[Optional[Dict[str, str]]]:
    return st.from_type(Optional[Dict[str, str]])  # type: ignore[arg-type]


def _lifecycle_strategy(
    field: Dataset.Field,
) -> st.SearchStrategy[Optional[model.Lifecycle]]:
    # Lifecycle contains fields that have `Any` types which `st.from_type` can't handle.
    return st.sampled_from((None, model.Lifecycle()))


_SPECIAL_FIELDS = {
    "contact_email": _email_field_strategy,
    "history": lambda field: st.none(),
    "job_parameters": _job_parameters_strategy,
    "lifecycle": _lifecycle_strategy,
    "owner_email": _email_field_strategy,
    "orcid_of_owner": _orcid_field_strategy,
    "meta": _scientific_metadata_strategy,
}

st.register_type_strategy(
    RemotePath,
    st.text(
        alphabet=string.ascii_lowercase + string.ascii_uppercase + string.digits + "/."
    ).map(RemotePath),
)

st.register_type_strategy(
    PID,
    st.builds(
        PID,
        prefix=st.text(alphabet=st.characters(blacklist_characters="/")) | st.none(),
        pid=st.text(),
    ),
)


def _field_strategy(field: Dataset.Field) -> st.SearchStrategy[Any]:
    if (strategy := _SPECIAL_FIELDS.get(field.name)) is not None:
        return strategy(field)

    typ = field.type if field.required else Optional[field.type]
    return st.from_type(typ)  # type:ignore[arg-type]


def _make_dataset(
    *, type: DatasetType, args: Dict[str, Any], read_only: Dict[str, Any]
) -> Dataset:
    dset = Dataset(type=type, **args)
    for key, val in read_only.items():
        setattr(dset, "_" + key, val)
    return dset


def datasets(
    type: Optional[DatasetType] = None, for_upload: bool = False, **fields: Any
) -> st.SearchStrategy[Dataset]:
    """A strategy for generating datasets.

    This strategy can populate all dataset fields.
    However, there are some limitations:

    - Some complex models may be uninitialized, e.g., ``lifecycle``.
    - Fields of type ``dict`` only have string values, e.g., ``meta`` will only
      be a ``dict[str, str]`` instead of the broader value types allowed by SciCat.
    - The dataset has no files.

    Parameters
    ----------
    type:
        The type of dataset to generate.
        If ``None``, a random dataset type will be chosen.
    for_upload:
        If ``True``, the dataset can be uploaded because only writable fields
        will be set.
        Otherwise, read-only fields may be set as well.
    fields:
        Concrete values or specific search strategies for dataset fields.

    Returns
    -------
    :
        Datasets.

    Examples
    --------
    To draw arbitrary datasets, use

    .. code-block:: python

        from hypothesis import given
        from scitacean.testing import strategies as sst

        @given(ds=sst.datasets())
        def test_dataset(ds):
            # use ds

    Limit to derived datasets that can be uploaded:

    .. code-block:: python

        @given(ds=sst.datasets(type="derived", for_upload=True))
        def test_derived_upload(ds):
            client = ...
            client.upload_new_dataset_now(ds)

    Fields can be fixed to specific values or generated from specific strategies.
    All other fields are generated as normal.

    .. code-block:: python

        @given(ds=sst.datasets(
            owner="librarian",
            owner_group=st.sampled_from(("library", "faculty"))
        ))
        def test_dataset_owner(ds):
            assert ds.owner == "librarian"
            assert ds.owner_group in ("library", "faculty")
            # other tests

    It is also possible to fix read-only fields:

    .. code-block:: python

        @given(ds=sst.datasets(pid=PID.parse("abcd-12")))
        def test_dataset_fixed_pid(ds):
            assert ds.pid.prefix is None
            # other tests
    """
    if type is None:
        return st.sampled_from(DatasetType).flatmap(
            partial(datasets, for_upload=for_upload, **fields)
        )

    def make_fixed_arg(key: str) -> st.SearchStrategy[Any]:
        val = fields[key]
        return val if isinstance(val, st.SearchStrategy) else st.just(val)

    def make_arg(field: Dataset.Field) -> st.SearchStrategy[Any]:
        if field.name in fields:
            return make_fixed_arg(field.name)
        return _field_strategy(field)

    def make_args(read_only: bool) -> Dict[str, st.SearchStrategy[Any]]:
        return {
            field.name: make_arg(field)
            for field in Dataset.fields(read_only=read_only, dataset_type=type)
            if field.name != "type"
        }

    args = st.fixed_dictionaries(
        {
            **make_args(read_only=False),
            "checksum_algorithm": st.sampled_from(("blake2b", "sha256")),
        }
    )
    if not for_upload:
        read_only = st.fixed_dictionaries(make_args(read_only=True))
    else:
        read_only = st.just({})
    return st.builds(
        _make_dataset,
        type=st.just(type),
        args=args,
        read_only=read_only,
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""HTML representations of datasets for Jupyter."""

import html
from typing import Any, Dict, Iterable, List, Optional

import pydantic

from ..dataset import Dataset
from ..model import DatasetType
from . import _resources
from ._common_html import Field, format_field_flag, format_type, format_value


def dataset_html_repr(dset: Dataset) -> str:
    template = _resources.dataset_repr_template()
    style_sheet = _resources.dataset_style()
    fields = _get_fields(dset)
    main_rows = "\n".join(_format_field(field) for field in fields if field.main)
    detail_rows = "\n".join(_format_field(field) for field in fields if not field.main)
    return template.substitute(
        style_sheet=style_sheet,
        dataset_type=_format_dataset_type(dset.type),
        file_info=_format_file_info(dset, archived=False),
        archived_file_info=_format_file_info(dset, archived=True),
        main_rows=main_rows,
        detail_rows=detail_rows,
        file_rows=_format_files(dset),
        metadata_rows=_format_metadata(dset),
    )


def _format_files(dset: Dataset) -> str:
    template = _resources.files_repr_template()
    return "\n".join(
        template.substitute(
            local_path=(
                '<span class="cean-empty-field">None</span>'
                if not file.is_on_local
                else html.escape(str(file.local_path))
            ),
            remote_path=(
                '<span class="cean-empty-field">None</span>'
                if not file.is_on_remote
                else html.escape(str(file.remote_path))
            ),
            size=_human_readable_size(file.size),
        )
        for file in dset.files
    )


def _format_metadata(dset: Dataset) -> str:
    template = _resources.metadata_template()
    return "\n".join(
        template.substitute(
            name=html.escape(str(name)),
            value=_format_metadata_value(value),
        )
        for name, value in dset.meta.items()
    )


def _format_metadata_value(value: Any) -> str:
    if _has_value_unit_encoding(value):
        if (unit := value.get("unit")) is not None:
            return f"{html.escape(str(value['value']))} [{html.escape(str(unit))}]"
        else:
            return html.escape(str(value["value"]))

    return html.escape(str(value))


# ESS uses these keys for all scientific metadata.
# If the metadata uses only these keys, format it more nicely.
# Otherwise, fall back to showing the item's str.
_VALUE_UNIT_KEYS = {"value", "unit", "valueSI", "unitSI"}


def _has_value_unit_encoding(meta_value: Any) -> bool:
    if (
        isinstance(meta_value, dict)
        and "value" in meta_value
        and not (meta_value.keys() - _VALUE_UNIT_KEYS)
    ):
        return True
    return False


def _format_field(field: Field) -> str:
    name = field.name
    flag = format_field_flag(field)
    value = (
        '<span class="cean-empty-field">None</span>'
        if field.value is None
        else format_value(field.value)
    )
    typ = format_type(field.type)
    description = html.escape(field.description)
    row_highlight = _row_highlight_classes(field)

    template = _resources.dataset_field_repr_template()
    return template.substitute(
        name=name,
        flag=flag,
        type=typ,
        value=value,
        description=description,
        extra_classes=row_highlight,
    )


_EXCLUDED_FIELDS = {
    "type",
    "number_of_files",
    "number_of_files_archived",
    "size",
    "packed_size",
    "meta",
}

_MAIN_FIELDS = {
    "pid",
    "source_folder",
    "name",
    "description",
    "creation_time",
    "proposal_id",
    "sample_id",
    "input_datasets",
}


def _get_fields(dset: Dataset) -> List[Field]:
    validation = _validate(dset)
    fields = [
        Field(
            name=field.name,
            value=getattr(dset, field.name, None),
            type=field.type,
            description=field.description,
            read_only=field.read_only,
            required=field.required,
            error=_check_error(field, validation),
            main=field.name in _MAIN_FIELDS,
        )
        for field in dset.fields()
        if field.name not in _EXCLUDED_FIELDS
        and (field.used_by(dset.type) or getattr(dset, field.name, None) is not None)
    ]
    return sorted(
        sorted(fields, key=lambda field: field.name),
        key=lambda field: not field.required,
    )


def _check_error(field: Dataset.Field, validation: Dict[str, str]) -> Optional[str]:
    field_spec = next(filter(lambda f: f.name == field.name, Dataset.fields()))
    return validation.get(field_spec.scicat_name, None)


def _validate(dset: Dataset) -> Dict[str, str]:
    def single_elem(xs: Iterable[Any]) -> Any:
        (x,) = xs
        return x

    try:
        dset.validate()
        return {}
    except pydantic.ValidationError as e:
        return {single_elem(error["loc"]): error["msg"] for error in e.errors()}


def _format_dataset_type(dataset_type: DatasetType) -> str:
    return "Raw" if dataset_type == DatasetType.RAW else "Derived"


def _format_file_info(dset: Dataset, archived: bool) -> str:
    if archived:
        n_files = dset.number_of_files_archived
        if n_files == 0:
            # Don't show archived files if there aren't any.
            # But always show regular files.
            return ""
        size = dset.packed_size
        name = "archived: "
    else:
        n_files = dset.number_of_files
        size = dset.size
        name = ""

    return (
        f'{name} {n_files} <span class="cean-file-info-size">'
        f"({_human_readable_size(size)})</span>"
    )


def _human_readable_size(size_in_bytes: int) -> str:
    for power, prefix in ((4, "T"), (3, "G"), (2, "M"), (1, "k")):
        n = 1024**power
        if size_in_bytes >= n:
            return f"{size_in_bytes/n:.2f} {prefix}iB"
    return f"{size_in_bytes} B"


def _row_highlight_classes(field: Field) -> str:
    if field.required and field.value is None:
        return "cean-missing-value"
    # Do not flag read-only fields with a value as errors.
    # Validation is geared towards uploading where such fields must be None.
    # But here, we don't want to flag downloaded datasets as bad because of this.
    if field.error and not (field.read_only and field.error.startswith("Extra inputs")):
        return "cean-error"
    return ""



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
# mypy: disable-error-code="arg-type"

import pytest
from dateutil.parser import parse as parse_date

from scitacean import Client
from scitacean.client import ScicatClient
from scitacean.model import (
    DatasetType,
    UploadDataFile,
    UploadDerivedDataset,
    UploadOrigDatablock,
)
from scitacean.testing.backend.seed import (
    INITIAL_ORIG_DATABLOCKS,
)


@pytest.fixture()
def scicat_client(client: Client) -> ScicatClient:
    return client.scicat


@pytest.fixture()
def derived_dataset(scicat_access):
    return UploadDerivedDataset(
        contactEmail="black.foess@dom.koelle",
        creationTime=parse_date("1995-11-11T11:11:11.000Z"),
        owner="bfoess",
        investigator="b.foess@dom.koelle",
        sourceFolder="/dom/platt",
        type=DatasetType.DERIVED,
        inputDatasets=[],
        usedSoftware=[],
        ownerGroup=scicat_access.user.group,
        accessGroups=["koelle"],
        numberOfFilesArchived=0,
    )


@pytest.fixture()
def orig_datablock(scicat_access):
    # NOTE the placeholder!
    return UploadOrigDatablock(
        size=9235,
        dataFileList=[
            UploadDataFile(
                path="data.nxs", size=9235, time=parse_date("2023-08-18T13:52:33.000Z")
            )
        ],
        datasetId="PLACEHOLDER",
        ownerGroup=scicat_access.user.group,
        accessGroups=["group1", "2nd_group"],
    )


@pytest.mark.parametrize("key", ("raw", "derived"))
def test_get_orig_datablock(scicat_client, key):
    dblock = INITIAL_ORIG_DATABLOCKS[key][0]
    downloaded = scicat_client.get_orig_datablocks(dblock.datasetId)
    assert downloaded == [dblock]


def test_create_first_orig_datablock(scicat_client, derived_dataset, orig_datablock):
    uploaded = scicat_client.create_dataset_model(derived_dataset)
    orig_datablock.datasetId = uploaded.pid
    scicat_client.create_orig_datablock(orig_datablock)
    downloaded = scicat_client.get_orig_datablocks(uploaded.pid)
    assert len(downloaded) == 1
    downloaded = downloaded[0]
    for key, expected in orig_datablock:
        # The database populates a number of fields that are orig_datablock in dset.
        # But we don't want to test those here as we don't want to test the database.
        if expected is not None and key != "dataFileList":
            assert expected == dict(downloaded)[key], f"key = {key}"
    for i in range(len(orig_datablock.dataFileList)):
        for key, expected in orig_datablock.dataFileList[i]:
            assert (
                expected == dict(downloaded.dataFileList[i])[key]
            ), f"i = {i}, key = {key}"



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
# mypy: disable-error-code="arg-type"

from copy import deepcopy

import pytest
from dateutil.parser import parse as parse_date

from scitacean import PID, Client, ScicatCommError, Thumbnail
from scitacean.client import ScicatClient
from scitacean.model import (
    Attachment,
    DatasetType,
    UploadAttachment,
    UploadDerivedDataset,
)
from scitacean.testing.backend.seed import (
    INITIAL_ATTACHMENTS,
    INITIAL_DATASETS,
)


@pytest.fixture()
def scicat_client(client: Client) -> ScicatClient:
    return client.scicat


@pytest.fixture()
def derived_dataset(scicat_access):
    return UploadDerivedDataset(
        contactEmail="black.foess@dom.koelle",
        creationTime=parse_date("1995-11-11T11:11:11.000Z"),
        owner="bfoess",
        investigator="b.foess@dom.koelle",
        sourceFolder="/dom/platt",
        type=DatasetType.DERIVED,
        inputDatasets=[],
        usedSoftware=[],
        ownerGroup=scicat_access.user.group,
        accessGroups=["koelle"],
        numberOfFilesArchived=0,
    )


@pytest.fixture()
def attachment(scicat_access):
    return UploadAttachment(
        caption="An attachment",
        thumbnail=Thumbnail(mime="image/png", data=b"9278c78a904jh"),
        ownerGroup=scicat_access.user.group,
        accessGroups=["group1", "2nd_group"],
    )


def compare_attachment_after_upload(uploaded, downloaded):
    for key, expected in uploaded:
        # The database populates a number of fields that are None in uploaded.
        # But we don't want to test those here as we don't want to test the database.
        if expected is not None:
            assert expected == dict(downloaded)[key], f"key = {key}"


def test_create_attachment_for_dataset(scicat_client, attachment, derived_dataset):
    attachment1 = deepcopy(attachment)
    attachment2 = deepcopy(attachment)
    attachment2.caption = "Another attachment"

    dataset_id = scicat_client.create_dataset_model(derived_dataset).pid

    finalized1 = scicat_client.create_attachment_for_dataset(
        attachment1, dataset_id=dataset_id
    )
    attachment1.datasetId = dataset_id
    compare_attachment_after_upload(attachment1, finalized1)

    finalized2 = scicat_client.create_attachment_for_dataset(
        attachment2, dataset_id=dataset_id
    )
    attachment2.datasetId = dataset_id
    compare_attachment_after_upload(attachment2, finalized2)


def test_create_attachment_for_dataset_with_existing_id(
    real_client, attachment, derived_dataset, require_scicat_backend
):
    scicat_client = real_client.scicat

    dataset_id = scicat_client.create_dataset_model(derived_dataset).pid
    attachment.id = "attachment-id"
    scicat_client.create_attachment_for_dataset(attachment, dataset_id=dataset_id)

    with pytest.raises(ScicatCommError):
        scicat_client.create_attachment_for_dataset(attachment, dataset_id=dataset_id)


def test_cannot_create_attachment_for_dataset_for_nonexistent_dataset(
    scicat_client, attachment
):
    with pytest.raises(ScicatCommError):
        scicat_client.create_attachment_for_dataset(
            attachment, dataset_id=PID(pid="nonexistent-id")
        )


def test_create_attachment_for_dataset_for_dataset_populates_ids(
    scicat_client, attachment, derived_dataset
):
    assert attachment.id is None
    assert attachment.datasetId is None
    assert attachment.sampleId is None
    assert attachment.proposalId is None

    dataset_id = scicat_client.create_dataset_model(derived_dataset).pid

    finalized = scicat_client.create_attachment_for_dataset(
        attachment, dataset_id=dataset_id
    )

    assert finalized.id is not None
    assert finalized.datasetId is not None
    assert finalized.sampleId is None
    assert finalized.proposalId is None


def test_get_attachments_for_dataset(scicat_client):
    dset = INITIAL_DATASETS["derived"]
    attachments = scicat_client.get_attachments_for_dataset(dset.pid)
    assert attachments == INITIAL_ATTACHMENTS["derived"]


def test_get_attachments_for_dataset_no_attachments(scicat_client):
    assert INITIAL_ATTACHMENTS.get("raw") is None
    dset = INITIAL_DATASETS["raw"]
    attachments = scicat_client.get_attachments_for_dataset(dset.pid)
    assert attachments == []


@pytest.mark.parametrize("key", ("raw", "derived"))
def test_get_dataset_does_not_initialise_attachments(client, key):
    dset = INITIAL_DATASETS["derived"]
    downloaded = client.get_dataset(dset.pid)
    assert downloaded.attachments is None


@pytest.mark.parametrize("key", ("raw", "derived"))
def test_download_attachments_for_dataset(client, key):
    dset = INITIAL_DATASETS[key]
    downloaded = client.get_dataset(dset.pid)
    with_attachments = client.download_attachments_for(downloaded)
    expected = [
        Attachment.from_download_model(attachment)
        for attachment in INITIAL_ATTACHMENTS.get(key, ())
    ]
    assert with_attachments.attachments == expected


@pytest.mark.parametrize("key", ("raw", "derived"))
def test_get_dataset_with_attachments(client, key):
    dset = INITIAL_DATASETS[key]
    downloaded = client.get_dataset(dset.pid, attachments=True)
    expected = [
        Attachment.from_download_model(attachment)
        for attachment in INITIAL_ATTACHMENTS.get(key, ())
    ]
    assert downloaded.attachments == expected



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Load schemas from a SciCat API."""

import dataclasses
from typing import Any, Dict, Optional

import requests


@dataclasses.dataclass
class SchemaField:
    name: str
    description: str
    type: str
    required: bool
    default: Optional[str]


@dataclasses.dataclass
class Schema:
    name: str
    fields: Dict[str, SchemaField]


def parse_field_type(spec: Dict[str, Any]):
    if "allOf" in spec:
        if len(spec["allOf"]) != 1:
            raise ValueError("More than one alternative type in 'allOf'")
        return parse_field_type(spec["allOf"][0])
    if "$ref" in spec:
        return spec["$ref"].rsplit("/", 1)[1]
    if "enum" in spec:
        if spec["type"] != "string":
            raise ValueError(f"Enum fields must have type 'string', got: {spec}")
        return "Enum[" + ", ".join(spec["enum"]) + "]"
    if spec["type"] == "number":
        return "int"
    if spec["type"] == "string":
        if spec.get("format", "") == "date-time":
            return "datetime"
        return "str"
    if spec["type"] == "boolean":
        return "bool"
    if spec["type"] == "array":
        return "List[{}]".format(parse_field_type(spec["items"]))
    if spec["type"] == "object":
        return "Dict[str, Any]"
    raise ValueError(f"Unknown field type: {spec['type']}")


def parse_schema_fields(spec: Dict[str, Any]) -> Dict[str, SchemaField]:
    return {
        name: SchemaField(
            name=name,
            description=field_spec.get("description", ""),
            type=parse_field_type(field_spec),
            required=name in spec.get("required", ()),
            default=field_spec.get("default", None),
        )
        for name, field_spec in spec["properties"].items()
    }


def parse_schemas(spec: Dict[str, Any]) -> Dict[str, Schema]:
    schemas = {}
    for name, field_spec in spec.items():
        print(f"Parsing schema {name}")  # noqa: T201
        schemas[name] = Schema(name=name, fields=parse_schema_fields(field_spec))
    return schemas


def fetch_specs(url: str) -> Dict[str, Any]:
    """Download the raw schema JSON from the API.

    ``url`` needs to point to a 'json-explorer' of a SciCat backend with version >= 4.
    E.g. ``http://localhost:3000/explorer-json`` for a locally running instance.
    """
    response = requests.get(url, timeout=10)
    if not response.ok:
        raise RuntimeError(
            f"Failed to fetch specs: {response.status_code} {response.reason}"
        )
    return response.json()


def load_schemas(url: str) -> Dict[str, Schema]:
    spec_json = fetch_specs(url)
    return parse_schemas(spec_json["components"]["schemas"])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Backend configuration."""

from dataclasses import dataclass
from typing import Dict


@dataclass
class SciCatUser:
    """A SciCat user.

    Warning
    -------
    Only ever use this for testing!
    This class does not have the usual protections against
    leaks of secrets used by the client.
    """

    username: str
    password: str
    email: str
    group: str

    @property
    def credentials(self) -> Dict[str, str]:
        """Return login credentials for this user.

        User as

        .. code-block:: python

            client = Client.from_credentials(url="...", **user.credentials)
        """
        return {
            "username": self.username,
            "password": self.password,
        }


# see https://github.com/SciCatProject/scicat-backend-next/blob/master/src/config/configuration.ts
USERS = {
    "ingestor": SciCatUser(
        username="ingestor",
        password="aman",  # noqa: S106
        email="scicatingestor@your.site",
        group="ingestor",
    ),
    "user1": SciCatUser(
        username="user1",
        password="a609316768619f154ef58db4d847b75e",  # noqa: S106
        email="user1@your.site",
        group="group1",
    ),
    "user2": SciCatUser(
        username="user2",
        password="f522d1d715970073a6413474ca0e0f63",  # noqa: S106
        email="user2@your.site",
        group="group2",
    ),
    "user3": SciCatUser(
        username="user3",
        password="70dc489e8ee823ae815e18d664424df2",  # noqa: S106
        email="user3@your.site",
        group="group3",
    ),
    "user4": SciCatUser(
        username="user4",
        password="0014890e7020f515b92b767227ef2dfa",  # noqa: S106
        email="user4@your.site",
        group="group4",
    ),
    "user5.1": SciCatUser(
        username="user5.1",
        password="359a5fda99bfe5dbc42ee9b3ede77fb7",  # noqa: S106
        email="user5.1@your.site",
        group="group5",
    ),
    "user5.2": SciCatUser(
        username="user5.2",
        password="f3ebd2e4def95db59ef95ee32ef45242",  # noqa: S106
        email="user5.2@your.site",
        group="group5",
    ),
}
"""Pre-configured users of the backend."""

SCICAT_PORT = 3000
"""Port of the SciCat server on localhost."""
PID_PREFIX = "PID.prefix.a0b1"
"""Prefix for PIDs that gets inserted by the server."""
SITE = "SCITACEAN"
"""Name of the deployment site (facility) of SciCat."""


@dataclass
class SciCatAccess:
    """Access parameters for a local SciCat backend."""

    url: str
    user: SciCatUser


def local_access(user: str) -> SciCatAccess:
    """Return parameters to connect a client to a local SciCat backend.

    Parameters
    ----------
    user:
        User name.
        Must be in :attr:`scitacean.testing.backend.config.USERS`.

    Returns
    -------
    :
        Parameters for the local SciCat backend.
    """
    return SciCatAccess(url=f"http://localhost:{SCICAT_PORT}/api/v3/", user=USERS[user])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Fake client for testing."""

from __future__ import annotations

import datetime
import functools
import uuid
from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from .. import model
from ..client import Client, ScicatClient
from ..error import ScicatCommError
from ..pid import PID
from ..typing import FileTransfer
from ..util.credentials import StrStorage


def _conditionally_disabled(func: Callable[..., Any]) -> Callable[..., Any]:
    @functools.wraps(func)
    def impl(self, *args, **kwargs):  # type: ignore[no-untyped-def]
        if (exc := self.main.disabled.get(func.__name__)) is not None:
            raise exc
        return func(self, *args, **kwargs)

    return impl


# Inherits from Client to satisfy type hints.
class FakeClient(Client):
    """Mimics a client without accessing the internet.

    This class is a stand in for :class:`scitacean.Client` to emulate downloading and
    uploading of datasets without actually accessing a SciCat server.

    Since this client fakes communication with SciCat, it stores raw
    `pydantic <https://docs.pydantic.dev/>`_ models, namely

    - ``FakeClient.datasets``:
            :class:`dict` of :class:`scitacean.model.DownloadDataset`,
            indexed by dataset PID.
    - ``FakeClient.orig_datablocks``:
            :class:`dict` of lists of :class:`scitacean.model.DownloadOrigDatablock`,
            indexed by the *dataset* ID.
    - ``FakeClient.attachments``:
            :class:`dict` of lists of :class:`scitacean.model.DownloadAttachment`,
            indexed by the *dataset* ID.

    Individual functions can be disabled (that is, made to raise an exception)
    using the ``disabled`` argument of the initializer.

    Important
    ---------
    ``FakeClient`` does not behave exactly like a ``Client`` connected to a real
    server as that would require reimplementing a large part of the SciCat backend.
    You should thus always also test your code against a (potentially locally deployed)
    real server.

    In particular, do not rely on specific error messages or the detailed settings
    in the datasets returned by ``FakeClient.upload_new_dataset_now``!

    Examples
    --------
    Set up a fake client for download:

    .. code-block:: python

        client = FakeClient()
        pid = PID(prefix="sample.prefix", pid="1234-5678-abcd")
        client.datasets[pid] = model.DownloadDataset(pid=pid, ...)
        client.orig_datablocks[pid] = [model.OrigDatablock(
            datasetId=str(pid),
            ...
        )]
        client.get_dataset(pid)

    Upload a dataset:

    .. code-block:: python

        client = FakeClient(file_transfer=FakeFileTransfer())
        finalized = client.upload_new_dataset_now(dset)

        # contains new dataset and datablock:
        client.datasets[finalized.pid]
        client.orig_datablocks[finalized.pid]

    Disable a method:

    .. code-block:: python

        client = FakeClient(
            disable={"create_dataset_model": RuntimeError("Upload failed")})
        # raises RuntimeError("Upload failed"):
        client.upload_new_dataset_now(dset)

    See Also
    --------
    scitacean.testing.transfer.FakeFileTransfer: Fake file up-/downloads
        without a real server.
    """

    def __init__(
        self,
        *,
        file_transfer: Optional[FileTransfer] = None,
        disable: Optional[Dict[str, Exception]] = None,
    ) -> None:
        """Initialize a fake client with empty dataset storage.

        Parameters
        ----------
        file_transfer:
            Typically :class:`scitacean.testing.transfer.FakeFileTransfer`
            but may be a real file transfer.
        disable:
            ``dict`` of function names to exceptions.
            Functions listed here raise the given exception
            when called and do nothing else.
        """
        super().__init__(client=FakeScicatClient(self), file_transfer=file_transfer)

        self.disabled = {} if disable is None else dict(disable)
        self.datasets: Dict[PID, model.DownloadDataset] = {}
        self.orig_datablocks: Dict[PID, List[model.DownloadOrigDatablock]] = {}
        self.attachments: Dict[PID, List[model.DownloadAttachment]] = {}

    @classmethod
    def from_token(
        cls,
        *,
        url: str,
        token: Union[str, StrStorage],
        file_transfer: Optional[FileTransfer] = None,
    ) -> FakeClient:
        """Create a new fake client.

        All arguments except ``file_Transfer`` are ignored.
        """
        return FakeClient(file_transfer=file_transfer)

    @classmethod
    def from_credentials(
        cls,
        *,
        url: str,
        username: Union[str, StrStorage],
        password: Union[str, StrStorage],
        file_transfer: Optional[FileTransfer] = None,
    ) -> FakeClient:
        """Create a new fake client.

        All arguments except ``file_Transfer`` are ignored.
        """
        return FakeClient(file_transfer=file_transfer)

    @classmethod
    def without_login(
        cls, *, url: str, file_transfer: Optional[FileTransfer] = None
    ) -> FakeClient:
        """Create a new fake client.

        All arguments except ``file_Transfer`` are ignored.
        """
        return FakeClient(file_transfer=file_transfer)


class FakeScicatClient(ScicatClient):
    """Mimics a ScicatClient, to be used by FakeClient."""

    def __init__(self, main_client: FakeClient) -> None:
        super().__init__(url="", token="", timeout=datetime.timedelta(seconds=60))
        self.main = main_client

    @_conditionally_disabled
    def get_dataset_model(
        self, pid: PID, strict_validation: bool = False
    ) -> model.DownloadDataset:
        """Fetch a dataset from SciCat."""
        _ = strict_validation  # unused by fake
        try:
            return self.main.datasets[pid]
        except KeyError:
            raise ScicatCommError(f"Unable to retrieve dataset {pid}") from None

    @_conditionally_disabled
    def get_orig_datablocks(
        self, pid: PID, strict_validation: bool = False
    ) -> List[model.DownloadOrigDatablock]:
        """Fetch an orig datablock from SciCat."""
        _ = strict_validation  # unused by fake
        try:
            return self.main.orig_datablocks[pid]
        except KeyError:
            raise ScicatCommError(
                f"Unable to retrieve orig datablock for dataset {pid}"
            ) from None

    @_conditionally_disabled
    def get_attachments_for_dataset(
        self, pid: PID, strict_validation: bool = False
    ) -> List[model.DownloadAttachment]:
        """Fetch all attachments from SciCat for a given dataset."""
        _ = strict_validation  # unused by fake
        return self.main.attachments.get(pid) or []

    @_conditionally_disabled
    def create_dataset_model(
        self, dset: Union[model.UploadDerivedDataset, model.UploadRawDataset]
    ) -> model.DownloadDataset:
        """Create a new dataset in SciCat."""
        ingested = _process_dataset(dset)
        pid: PID = ingested.pid  # type: ignore[assignment]
        self.main.datasets[pid] = ingested
        return ingested

    @_conditionally_disabled
    def create_orig_datablock(
        self, dblock: model.UploadOrigDatablock
    ) -> model.DownloadOrigDatablock:
        """Create a new orig datablock in SciCat."""
        ingested = _process_orig_datablock(dblock)
        dataset_id = ingested.datasetId
        if dataset_id not in self.main.datasets:
            raise ScicatCommError(f"No dataset with id {dataset_id}")
        self.main.orig_datablocks.setdefault(dataset_id, []).append(ingested)
        return ingested

    @_conditionally_disabled
    def create_attachment_for_dataset(
        self,
        attachment: model.UploadAttachment,
        dataset_id: Optional[PID] = None,
    ) -> model.DownloadAttachment:
        """Create a new attachment for a dataset in SciCat."""
        if dataset_id is None:
            dataset_id = attachment.datasetId
        if dataset_id is None:
            raise ValueError(
                "Cannot determine the dataset ID for attachment. "
                "Specify the ID as a function argument or in the attachment."
            )

        ingested = _process_attachment(attachment, dataset_id)
        if dataset_id not in self.main.datasets:
            raise ScicatCommError(f"No dataset with id {dataset_id}")
        self.main.attachments.setdefault(dataset_id, []).append(ingested)
        return ingested

    @_conditionally_disabled
    def validate_dataset_model(
        self, dset: Union[model.UploadDerivedDataset, model.UploadRawDataset]
    ) -> None:
        """Validate model remotely in SciCat."""
        # Models were locally validated on construction, assume they are valid.
        pass


def _model_dict(mod: model.BaseModel) -> Dict[str, Any]:
    return {
        key: deepcopy(val)
        for key in mod.get_model_fields().keys()
        if (val := getattr(mod, key)) is not None
    }


def _process_relationship(
    relationship: model.UploadRelationship,
) -> model.DownloadRelationship:
    return model.DownloadRelationship(**_model_dict(relationship))


def _process_technique(technique: model.UploadTechnique) -> model.DownloadTechnique:
    return model.DownloadTechnique(**_model_dict(technique))


def _process_data_file(file: model.UploadDataFile) -> model.DownloadDataFile:
    return model.DownloadDataFile(**_model_dict(file))


def _process_dataset(
    dset: Union[model.UploadDerivedDataset, model.UploadRawDataset]
) -> model.DownloadDataset:
    created_at = datetime.datetime.now(tz=datetime.timezone.utc)
    # TODO use user login if possible
    # Using strict_validation=False because the input model should already be validated.
    # If there are validation errors, it was probably intended by the user.
    fields = _model_dict(dset)
    if "relationships" in fields:
        fields["relationships"] = list(
            map(_process_relationship, fields["relationships"])
        )
    if "techniques" in fields:
        fields["techniques"] = list(map(_process_technique, fields["techniques"]))
    return model.construct(
        model.DownloadDataset,
        _strict_validation=False,
        pid=PID.generate(prefix="PID.prefix.a0b1"),
        createdBy="fake",
        createdAt=created_at,
        updatedBy="fake",
        updatedAt=created_at,
        **fields,
    )


def _process_orig_datablock(
    dblock: model.UploadOrigDatablock,
) -> model.DownloadOrigDatablock:
    created_at = datetime.datetime.now(tz=datetime.timezone.utc)
    # TODO use user login if possible
    # TODO more fields
    # Using strict_validation=False because the input model should already be validated.
    # If there are validation errors, it was probably intended by the user.
    fields = _model_dict(dblock)
    if "dataFileList" in fields:
        fields["dataFileList"] = list(map(_process_data_file, fields["dataFileList"]))
    processed = model.construct(
        model.DownloadOrigDatablock,
        _strict_validation=False,
        createdBy="fake",
        createdAt=created_at,
        updatedBy="fake",
        updatedAt=created_at,
        **fields,
    )
    if dblock.datasetId is not None:
        processed.datasetId = dblock.datasetId
    return processed


def _process_attachment(
    attachment: model.UploadAttachment, dataset_id: Optional[PID] = None
) -> model.DownloadAttachment:
    created_at = datetime.datetime.now(tz=datetime.timezone.utc)
    # Using strict_validation=False because the input model should already be validated.
    # If there are validation errors, it was probably intended by the user.
    fields = _model_dict(attachment)
    if fields.get("id") is None:
        fields["id"] = uuid.uuid4().hex
    if dataset_id is not None:
        fields["datasetId"] = dataset_id
    return model.construct(
        model.DownloadAttachment,
        _strict_validation=False,
        createdBy="fake",
        createdAt=created_at,
        updatedBy="fake",
        updatedAt=created_at,
        **fields,
    )


def process_uploaded_dataset(
    dataset: Union[model.UploadDerivedDataset, model.UploadRawDataset],
    orig_datablocks: Optional[List[model.UploadOrigDatablock]],
    attachments: Optional[List[model.UploadAttachment]],
) -> Tuple[
    model.DownloadDataset,
    Optional[List[model.DownloadOrigDatablock]],
    Optional[List[model.DownloadAttachment]],
]:
    """Process a dataset as if it was uploaded to SciCat.

    This function attempts to mimic how SciCat converts uploaded dataset
    (and associated) models to the in-database (and download) models.
    It is not completely faithful to the real SciCat but only an approximation.
    """
    dblocks = (
        list(map(_process_orig_datablock, orig_datablocks))
        if orig_datablocks is not None
        else None
    )
    atts = (
        list(map(_process_attachment, attachments)) if attachments is not None else None
    )
    return _process_dataset(dataset), dblocks, atts



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""SFTP file transfer."""

import os
from contextlib import contextmanager
from datetime import datetime, timezone
from pathlib import Path
from typing import Callable, Iterator, List, Optional, Union

from paramiko import SFTPAttributes, SFTPClient, SSHClient

from ..dataset import Dataset
from ..error import FileUploadError
from ..file import File
from ..filesystem import RemotePath
from ..logging import get_logger
from .util import source_folder_for


class SFTPDownloadConnection:
    """Connection for downloading files with SFTP.

    Should be created using
    :meth:`scitacean.transfer.sftp.SFTPFileTransfer.connect_for_download`.
    """

    def __init__(self, *, sftp_client: SFTPClient, host: str) -> None:
        self._sftp_client = sftp_client
        self._host = host

    def download_files(self, *, remote: List[RemotePath], local: List[Path]) -> None:
        """Download files from the given remote path."""
        for r, l in zip(remote, local):
            self.download_file(remote=r, local=l)

    def download_file(self, *, remote: RemotePath, local: Path) -> None:
        """Download a file from the given remote path."""
        get_logger().info(
            "Downloading file %s from host %s to %s",
            remote,
            self._host,
            local,
        )
        self._sftp_client.get(remotepath=remote.posix, localpath=os.fspath(local))


class SFTPUploadConnection:
    """Connection for uploading files with SFTP.

    Should be created using
    :meth:`scitacean.transfer.sftp.SFTPFileTransfer.connect_for_upload`.
    """

    def __init__(
        self, *, sftp_client: SFTPClient, source_folder: RemotePath, host: str
    ) -> None:
        self._sftp_client = sftp_client
        self._source_folder = source_folder
        self._host = host

    @property
    def source_folder(self) -> RemotePath:
        """The source folder this connection uploads to."""
        return self._source_folder

    def remote_path(self, filename: Union[str, RemotePath]) -> RemotePath:
        """Return the complete remote path for a given path."""
        return self.source_folder / filename

    def _make_source_folder(self) -> None:
        try:
            _mkdir_remote(self._sftp_client, self.source_folder)
        except OSError as exc:
            raise FileUploadError(
                f"Failed to create source folder {self.source_folder}: {exc.args}"
            ) from None

    def upload_files(self, *files: File) -> List[File]:
        """Upload files to the remote folder."""
        self._make_source_folder()
        uploaded: list[File] = []
        try:
            uploaded.extend(self._upload_file(file) for file in files)
        except Exception:
            self.revert_upload(*uploaded)
            raise
        return uploaded

    def _upload_file(self, file: File) -> File:
        if file.local_path is None:
            raise ValueError(
                f"Cannot upload file to {file.remote_path}, "
                "the file has no local path"
            )
        remote_path = self.remote_path(file.remote_path)
        get_logger().info(
            "Uploading file %s to %s on host %s",
            file.local_path,
            remote_path,
            self._host,
        )
        st = self._sftp_client.put(
            remotepath=remote_path.posix, localpath=os.fspath(file.local_path)
        )
        return file.uploaded(
            remote_gid=str(st.st_gid),
            remote_uid=str(st.st_uid),
            remote_creation_time=datetime.now().astimezone(timezone.utc),
            remote_perm=str(st.st_mode),
            remote_size=st.st_size,
        )

    def revert_upload(self, *files: File) -> None:
        """Remove uploaded files from the remote folder."""
        for file in files:
            self._revert_upload_single(remote=file.remote_path, local=file.local_path)

        if _remote_folder_is_empty(self._sftp_client, self.source_folder):
            try:
                get_logger().info(
                    "Removing empty remote directory %s on host %s",
                    self.source_folder,
                    self._host,
                )
                self._sftp_client.rmdir(self.source_folder.posix)
            except IOError as exc:
                get_logger().warning(
                    "Failed to remove empty remote directory %s on host %s:\n%s",
                    self.source_folder,
                    self._host,
                    exc,
                )

    def _revert_upload_single(
        self, *, remote: RemotePath, local: Optional[Path]
    ) -> None:
        remote_path = self.remote_path(remote)
        get_logger().info(
            "Reverting upload of file %s to %s on host %s",
            local,
            remote_path,
            self._host,
        )

        try:
            self._sftp_client.remove(remote_path.posix)
        except IOError as exc:
            get_logger().warning("Error reverting file %s:\n%s", remote_path, exc)
            return


class SFTPFileTransfer:
    """Upload / download files using SFTP.

    Configuration & Authentication
    ------------------------------
    The file transfer connects to the server at the address given
    as the ``host`` constructor argument.
    This may be

    - a full url such as ``some.fileserver.edu``,
    - or an IP address like ``127.0.0.1``.

    The file transfer can currently only authenticate through an SSH agent.
    The agent must be set up for the chosen host and hold a valid key.
    If this is not the case, it is possible to inject a custom ``connect`` function
    that authenticates in a different way.
    See the examples below.

    Upload folder
    -------------
    The file transfer can take an optional ``source_folder`` as a constructor argument.
    If it is given, ``SFTPFileTransfer`` uploads all files to it and ignores the
    source folder set in the dataset.
    If it is not given, ``SFTPFileTransfer`` uses the dataset's source folder.

    The source folder argument to ``SFTPFileTransfer`` may be a Python format string.
    In that case, all format fields are replaced by the corresponding fields
    of the dataset.
    All non-ASCII characters and most special ASCII characters are replaced.
    This should avoid broken paths from essentially random contents in datasets.

    Examples
    --------
    Given

    .. code-block:: python

        dset = Dataset(type="raw",
                       name="my-dataset",
                       source_folder="/dataset/source",
                       )

    This uploads to ``/dataset/source``:

    .. code-block:: python

        file_transfer = SFTPFileTransfer(host="fileserver")

    This uploads to ``/transfer/folder``:

    .. code-block:: python

        file_transfer = SFTPFileTransfer(host="fileserver",
                                         source_folder="transfer/folder")

    This uploads to ``/transfer/my-dataset``:
    (Note that ``{name}`` is replaced by ``dset.name``.)

    .. code-block:: python

        file_transfer = SFTPFileTransfer(host="fileserver",
                                        source_folder="transfer/{name}")

    A useful approach is to include a unique ID in the source folder, for example,
    ``"/some/base/folder/{uid}"``, to avoid clashes between different datasets.
    Scitacean will fill in the ``"{uid}"`` placeholder with a new UUID4.

    The connection and authentication method can be customized
    using the ``connect`` argument.
    For example, to use a specific username + SSH key file, use the following:

    .. code-block:: python

        def connect(host, port):
            from paramiko import SSHClient

            client = SSHClient()
            client.load_system_host_keys()
            client.connect(
                hostname=host,
                port=port,
                username="<username>",
                key_filename="<key-file-name>",
            )
            return client.open_sftp()

        file_transfer = SFTPFileTransfer(host="fileserver", connect=connect)

    The :class:`paramiko.client.SSHClient` can be configured as needed in this function.
    """

    def __init__(
        self,
        *,
        host: str,
        port: int = 22,
        source_folder: Optional[Union[str, RemotePath]] = None,
        connect: Optional[Callable[[str, Optional[int]], SFTPClient]] = None,
    ) -> None:
        """Construct a new SFTP file transfer.

        Parameters
        ----------
        host:
            URL or name of the server to connect to.
        port:
            Port of the server.
        source_folder:
            Upload files to this folder if set.
            Otherwise, upload to the dataset's source_folder.
            Ignored when downloading files.
        connect:
            If this argument is set, it will be called to create a client
            for the server instead of the builtin method.
            The function arguments are ``host`` and ``port`` as determined by the
            arguments to ``__init__`` shown above.
        """
        self._host = host
        self._port = port
        self._source_folder_pattern = (
            RemotePath(source_folder)
            if isinstance(source_folder, str)
            else source_folder
        )
        self._connect = connect

    def source_folder_for(self, dataset: Dataset) -> RemotePath:
        """Return the source folder used for the given dataset."""
        return source_folder_for(dataset, self._source_folder_pattern)

    @contextmanager
    def connect_for_download(self) -> Iterator[SFTPDownloadConnection]:
        """Create a connection for downloads, use as a context manager."""
        sftp_client = _connect(self._host, self._port, connect=self._connect)
        try:
            yield SFTPDownloadConnection(sftp_client=sftp_client, host=self._host)
        finally:
            sftp_client.close()

    @contextmanager
    def connect_for_upload(self, dataset: Dataset) -> Iterator[SFTPUploadConnection]:
        """Create a connection for uploads, use as a context manager.

        Parameters
        ----------
        dataset:
            The connection will be used to upload files of this dataset.
            Used to determine the target folder.
        """
        source_folder = self.source_folder_for(dataset)
        sftp_client = _connect(self._host, self._port, connect=self._connect)
        try:
            yield SFTPUploadConnection(
                sftp_client=sftp_client, source_folder=source_folder, host=self._host
            )
        finally:
            sftp_client.close()


def _default_connect(host: str, port: Optional[int]) -> SFTPClient:
    client = SSHClient()
    client.load_system_host_keys()
    if port is not None:
        client.connect(hostname=host, port=port)
    else:
        client.connect(hostname=host)
    return client.open_sftp()


def _connect(
    host: str,
    port: int,
    connect: Optional[Callable[[str, Optional[int]], SFTPClient]],
) -> SFTPClient:
    try:
        if connect is None:
            connect = _default_connect
        return connect(host, port)
    except Exception as exc:
        # We pass secrets as arguments to functions called in this block, and those
        # can be leaked through exception handlers. So catch all exceptions
        # and strip the backtrace up to this point to hide those secrets.
        raise type(exc)(exc.args) from None


def _remote_folder_is_empty(sftp: SFTPClient, path: RemotePath) -> bool:
    return not sftp.listdir(path.posix)


def _mkdir_remote(sftp: SFTPClient, path: RemotePath) -> None:
    if (parent := path.parent) not in (".", "/"):
        _mkdir_remote(sftp, parent)

    st_stat = _try_remote_stat(sftp, path)
    if st_stat is None:
        sftp.mkdir(path.posix)
    elif not _is_remote_dir(st_stat):
        raise FileExistsError(
            f"Cannot make directory because path points to a file: {path}"
        )


def _try_remote_stat(sftp: SFTPClient, path: RemotePath) -> Optional[SFTPAttributes]:
    try:
        return sftp.stat(path.posix)
    except FileNotFoundError:
        return None


def _is_remote_dir(st_stat: SFTPAttributes) -> bool:
    if st_stat.st_mode is None:
        return True  # Assume it is a dir and let downstream code fail if it isn't.
    return st_stat.st_mode & 0o040000 == 0o040000


__all__ = ["SFTPFileTransfer", "SFTPUploadConnection", "SFTPDownloadConnection"]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
# mypy: disable-error-code="arg-type, union-attr"

from datetime import datetime, timedelta, timezone
from pathlib import Path

import pytest
from dateutil.parser import parse as parse_datetime
from hypothesis import assume, given, settings
from hypothesis import strategies as st

from scitacean import PID, Dataset, DatasetType, File, RemotePath, model
from scitacean.testing import strategies as sst
from scitacean.testing.client import process_uploaded_dataset

from .common.files import make_file


@pytest.fixture()
def raw_download_model():
    return model.DownloadDataset(
        contactEmail="p.stibbons@uu.am",
        creationLocation="UnseenUniversity",
        creationTime=parse_datetime("1995-08-06T14:14:14Z"),
        inputDatasets=None,
        investigator=None,
        numberOfFilesArchived=None,
        owner="pstibbons",
        ownerGroup="faculty",
        principalInvestigator="Ponder Stibbons",
        sourceFolder=RemotePath("/uu/hex"),
        type=DatasetType.RAW,
        usedSoftware=None,
        accessGroups=["uu"],
        version="3",
        classification="IN=medium,AV=low,CO=low",
        comment="Where did this come from?",
        createdAt=parse_datetime("1995-08-06T14:14:14Z"),
        createdBy="pstibbons",
        dataFormat=".thaum",
        dataQualityMetrics=24,
        description="Some shady data",
        endTime=parse_datetime("1995-08-03T00:00:00Z"),
        history=None,
        instrumentGroup="professors",
        instrumentId="0000-aa",
        isPublished=True,
        jobLogData=None,
        jobParameters=None,
        keywords=["thaum", "shady"],
        license="NoThouchy",
        datasetName="Flux44",
        numberOfFiles=1,
        orcidOfOwner="https://orcid.org/0000-0001-2345-6789",
        ownerEmail="m.ridcully@uu.am",
        packedSize=0,
        pid=PID.parse("123.cc/948.f7.2a"),
        proposalId="33.dc",
        sampleId="bac.a4",
        sharedWith=["librarian"],
        size=400,
        sourceFolderHost="ftp://uu.am/data",
        updatedAt=parse_datetime("1995-08-06T17:30:18Z"),
        updatedBy="librarian",
        validationStatus="ok",
        datasetlifecycle=model.DownloadLifecycle(
            isOnCentralDisk=True,
            retrievable=False,
        ),
        relationships=[
            model.DownloadRelationship(
                pid=PID.parse("123.cc/020.0a.4e"),
                relationship="calibration",
            )
        ],
        techniques=[
            model.DownloadTechnique(
                name="reflorbment",
                pid="aa.90.4",
            )
        ],
        scientificMetadata={
            "confidence": "low",
            "price": {"value": "606", "unit": "AM$"},
        },
    )


@pytest.fixture()
def derived_download_model():
    return model.DownloadDataset(
        contactEmail="p.stibbons@uu.am",
        creationLocation=None,
        creationTime=parse_datetime("1995-08-06T14:14:14Z"),
        inputDatasets=[PID.parse("123.cc/948.f7.2a")],
        investigator="Ponder Stibbons",
        numberOfFilesArchived=None,
        owner="pstibbons",
        ownerGroup="faculty",
        principalInvestigator=None,
        sourceFolder=RemotePath("/uu/hex"),
        type=DatasetType.DERIVED,
        usedSoftware=["scitacean"],
        accessGroups=["uu"],
        version="3",
        classification="IN=medium,AV=low,CO=low",
        comment="Why did we actually make this data?",
        createdAt=parse_datetime("1995-08-06T14:14:14Z"),
        createdBy="pstibbons",
        dataFormat=None,
        dataQualityMetrics=24,
        description="Dubiously analyzed data",
        endTime=None,
        history=None,
        instrumentGroup="professors",
        instrumentId=None,
        isPublished=True,
        jobLogData="process interrupted",
        jobParameters={"nodes": 4},
        keywords=["thaum", "dubious"],
        license="NoThouchy",
        datasetName="Flux peaks",
        numberOfFiles=1,
        orcidOfOwner="https://orcid.org/0000-0001-2345-6789",
        ownerEmail="m.ridcully@uu.am",
        packedSize=0,
        pid=PID.parse("123.cc/948.f7.2a"),
        proposalId=None,
        sampleId=None,
        sharedWith=["librarian"],
        size=400,
        sourceFolderHost="ftp://uu.am/data",
        updatedAt=parse_datetime("1995-08-06T17:30:18Z"),
        updatedBy="librarian",
        validationStatus="ok",
        datasetlifecycle=model.DownloadLifecycle(
            isOnCentralDisk=True,
            retrievable=False,
        ),
        relationships=[
            model.DownloadRelationship(
                pid=PID.parse("123.cc/020.0a.4e"),
                relationship="calibration",
            )
        ],
        techniques=[
            model.DownloadTechnique(
                name="reflorbment",
                pid="aa.90.4",
            )
        ],
        scientificMetadata={
            "confidence": "low",
            "price": {"value": "606", "unit": "AM$"},
        },
    )


@pytest.fixture(params=["raw_download_model", "derived_download_model"])
def dataset_download_model(request):
    return request.getfixturevalue(request.param)


def test_from_download_models_initializes_fields(dataset_download_model):
    def get_model_field(name):
        val = getattr(dataset_download_model, name)
        if name == "relationships":
            return [model.Relationship.from_download_model(v) for v in val]
        if name == "techniques":
            return [model.Technique.from_download_model(v) for v in val]
        if name == "datasetlifecycle":
            return model.Lifecycle.from_download_model(val)
        return val

    dset = Dataset.from_download_models(dataset_download_model, [])
    for field in dset.fields():
        if field.used_by(dataset_download_model.type):
            assert getattr(dset, field.name) == get_model_field(field.scicat_name)


def test_from_download_models_does_not_initialize_wrong_fields(dataset_download_model):
    dset = Dataset.from_download_models(dataset_download_model, [])
    for field in dset.fields():
        if not field.used_by(dataset_download_model.type):
            assert getattr(dset, field.name) is None


@pytest.mark.parametrize("typ", (DatasetType.RAW, DatasetType.DERIVED))
def test_new_dataset_has_no_files(typ):
    dset = Dataset(type=typ)
    assert len(list(dset.files)) == 0
    assert dset.number_of_files == 0
    assert dset.number_of_files_archived == 0
    assert dset.packed_size == 0
    assert dset.size == 0


@pytest.mark.parametrize("typ", (DatasetType.RAW, DatasetType.DERIVED))
def test_add_local_file_to_new_dataset(typ, fs):
    file_data = make_file(fs, "local/folder/data.dat")

    dset = Dataset(type=typ)
    dset.add_local_files("local/folder/data.dat")

    assert len(list(dset.files)) == 1
    assert dset.number_of_files == 1
    assert dset.number_of_files_archived == 0
    assert dset.packed_size == 0
    assert dset.size == file_data["size"]

    [f] = dset.files
    assert not f.is_on_remote
    assert f.is_on_local
    assert f.remote_access_path(dset.source_folder) is None
    assert f.local_path == Path("local/folder/data.dat")
    assert f.size == file_data["size"]
    assert f.checksum_algorithm == "blake2b"

    assert abs(file_data["creation_time"] - f.creation_time) < timedelta(seconds=1)
    assert abs(file_data["creation_time"] - f.make_model().time) < timedelta(seconds=1)


@pytest.mark.parametrize("typ", (DatasetType.RAW, DatasetType.DERIVED))
def test_add_multiple_local_files_to_new_dataset(typ, fs):
    file_data0 = make_file(fs, "common/location1/data.dat")
    file_data1 = make_file(fs, "common/song.mp3")

    dset = Dataset(type=typ)
    dset.add_local_files("common/location1/data.dat", "common/song.mp3")

    assert len(list(dset.files)) == 2
    assert dset.number_of_files == 2
    assert dset.number_of_files_archived == 0
    assert dset.packed_size == 0
    assert dset.size == file_data0["size"] + file_data1["size"]

    [f0, f1] = dset.files
    if f0.local_path.suffix == ".mp3":
        f1, f0 = f0, f1

    assert not f0.is_on_remote
    assert f0.is_on_local
    assert f0.remote_access_path(dset.source_folder) is None
    assert f0.local_path == Path("common/location1/data.dat")
    assert f0.size == file_data0["size"]
    assert f0.checksum_algorithm == "blake2b"

    assert not f1.is_on_remote
    assert f1.is_on_local
    assert f1.remote_access_path(dset.source_folder) is None
    assert f1.local_path == Path("common/song.mp3")
    assert f1.size == file_data1["size"]
    assert f1.checksum_algorithm == "blake2b"


@pytest.mark.parametrize("typ", (DatasetType.RAW, DatasetType.DERIVED))
def test_add_multiple_local_files_to_new_dataset_with_base_path(typ, fs):
    file_data0 = make_file(fs, "common/location1/data.dat")
    file_data1 = make_file(fs, "common/song.mp3")

    dset = Dataset(type=typ)
    dset.add_local_files(
        "common/location1/data.dat", "common/song.mp3", base_path="common"
    )

    assert len(list(dset.files)) == 2
    assert dset.number_of_files == 2
    assert dset.number_of_files_archived == 0
    assert dset.packed_size == 0
    assert dset.size == file_data0["size"] + file_data1["size"]

    [f0, f1] = dset.files
    if f0.local_path.suffix == ".mp3":
        f1, f0 = f0, f1

    assert not f0.is_on_remote
    assert f0.is_on_local
    assert f0.remote_access_path(dset.source_folder) is None
    assert f0.local_path == Path("common/location1/data.dat")
    assert f0.size == file_data0["size"]
    assert f0.checksum_algorithm == "blake2b"

    assert not f1.is_on_remote
    assert f1.is_on_local
    assert f1.remote_access_path(dset.source_folder) is None
    assert f1.local_path == Path("common/song.mp3")
    assert f1.size == file_data1["size"]
    assert f1.checksum_algorithm == "blake2b"


@pytest.mark.parametrize("typ", (DatasetType.RAW, DatasetType.DERIVED))
@pytest.mark.parametrize("algorithm", ("sha256", None))
def test_can_set_default_checksum_algorithm(typ, algorithm, fs):
    make_file(fs, "local/data.dat")

    dset = Dataset(type=typ, checksum_algorithm=algorithm)
    dset.add_local_files("local/data.dat")

    [f] = dset.files
    assert f.checksum_algorithm == algorithm


@given(sst.datasets(for_upload=True))
@settings(max_examples=100)
def test_dataset_models_roundtrip(initial):
    dataset_model = initial.make_upload_model()
    dblock_models = initial.make_datablock_upload_models().orig_datablocks
    attachment_models = initial.make_attachment_upload_models()
    dataset_model, dblock_models, attachment_models = process_uploaded_dataset(
        dataset_model, dblock_models, attachment_models
    )
    dataset_model.createdAt = None
    dataset_model.createdBy = None
    dataset_model.updatedAt = None
    dataset_model.updatedBy = None
    dataset_model.pid = None
    rebuilt = Dataset.from_download_models(
        dataset_model=dataset_model,
        orig_datablock_models=dblock_models,
        attachment_models=attachment_models,
    )
    assert initial == rebuilt


@given(sst.datasets())
@settings(max_examples=10)
def test_make_scicat_models_datablock_without_files(dataset):
    assert dataset.make_datablock_upload_models().orig_datablocks is None


@given(sst.datasets(pid=st.builds(PID)))
@settings(max_examples=10)
def test_make_scicat_models_datablock_with_one_file(dataset):
    file_model = model.DownloadDataFile(
        path="path", size=6163, chk="8450ac0", gid="group", time=datetime.now()
    )
    dataset.add_files(File.from_download_model(local_path=None, model=file_model))

    blocks = dataset.make_datablock_upload_models().orig_datablocks
    assert len(blocks) == 1

    block = blocks[0]
    assert block.size == 6163
    assert block.datasetId == dataset.pid
    assert block.dataFileList == [model.UploadDataFile(**file_model.model_dump())]


def test_attachments_are_empty_by_default():
    dataset = Dataset(
        type="raw",
        owner="ridcully",
    )
    assert dataset.attachments == []


def test_attachments_are_none_after_from_download_models(dataset_download_model):
    dataset = Dataset.from_download_models(dataset_download_model, [])
    assert dataset.attachments is None


def test_attachments_initialized_in_from_download_models(dataset_download_model):
    dataset = Dataset.from_download_models(
        dataset_download_model,
        [],
        attachment_models=[
            model.DownloadAttachment(
                caption="The attachment",
                ownerGroup="att-owner",
            )
        ],
    )

    assert dataset.attachments == [
        model.Attachment(
            caption="The attachment",
            owner_group="att-owner",
        )
    ]


def test_can_add_attachment():
    dataset = Dataset(type="raw", owner_group="dset-owner")
    dataset.attachments.append(
        model.Attachment(
            caption="The attachment",
            owner_group="att-owner",
        )
    )
    assert dataset.attachments == [
        model.Attachment(
            caption="The attachment",
            owner_group="att-owner",
        )
    ]


def test_can_assign_attachments():
    dataset = Dataset(type="derived", owner_group="dset-owner")

    dataset.attachments = [
        model.Attachment(
            caption="The attachment",
            owner_group="att-owner",
        )
    ]
    assert dataset.attachments == [
        model.Attachment(
            caption="The attachment",
            owner_group="att-owner",
        )
    ]

    dataset.attachments = [
        model.Attachment(
            caption="Attachment 2",
            owner_group="other_owner",
        )
    ]
    assert dataset.attachments == [
        model.Attachment(
            caption="Attachment 2",
            owner_group="other_owner",
        )
    ]


def test_make_attachment_upload_models_fails_when_attachments_are_none():
    dataset = Dataset(type="derived", owner_group="dset-owner")
    dataset.attachments = None
    with pytest.raises(ValueError, match="attachment"):
        dataset.make_attachment_upload_models()


@given(sst.datasets())
@settings(max_examples=10)
def test_eq_self(dset):
    dset.add_files(
        File.from_download_model(
            local_path=None,
            model=model.DownloadDataFile(path="path", size=94571, time=datetime.now()),
        )
    )
    dset.attachments.append(
        model.Attachment(caption="The attachment", owner_group="owner")
    )
    assert dset == dset


# Fields whose types are not supported by hypothesis.
# E.g. because they contain `Any`.
_UNGENERATABLE_FIELDS = ("job_parameters", "meta")


# Filtering out bools because hypothesis struggles to find enough examples
# where the new value is not the same as the old value.
@pytest.mark.parametrize(
    "field",
    (
        f
        for f in Dataset.fields(read_only=False)
        if f.type != bool and f.name not in _UNGENERATABLE_FIELDS
    ),
    ids=lambda f: f.name,
)
@given(sst.datasets(), st.data())
@settings(max_examples=10)
def test_neq_single_mismatched_field_writable(field, initial, data):
    new_val = data.draw(st.from_type(field.type))
    assume(new_val != getattr(initial, field.name))
    modified = initial.replace(**{field.name: new_val})
    assert modified != initial
    assert initial != modified


@given(sst.datasets())
@settings(max_examples=10)
def test_neq_single_mismatched_file(initial):
    modified = initial.replace()
    modified.add_files(
        File.from_download_model(
            local_path=None,
            model=model.DownloadDataFile(
                path="path", size=51553312, time=datetime.now()
            ),
        )
    )
    initial.add_files(
        File.from_download_model(
            local_path=None,
            model=model.DownloadDataFile(path="path", size=94571, time=datetime.now()),
        )
    )
    assert modified != initial
    assert initial != modified


@given(sst.datasets())
@settings(max_examples=10)
def test_neq_extra_file(initial):
    modified = initial.replace()
    modified.add_files(
        File.from_download_model(
            local_path="/local",
            model=model.DownloadDataFile(
                path="path", size=51553312, time=datetime.now()
            ),
        )
    )
    assert modified != initial
    assert initial != modified


@given(sst.datasets())
@settings(max_examples=1)
def test_neq_attachment_none_vs_empty(initial):
    initial.attachments = []
    modified = initial.replace()
    modified.attachments = None
    assert initial != modified
    assert modified != initial


@given(sst.datasets())
@settings(max_examples=1)
def test_neq_extra_attachment(initial):
    initial.attachments = []
    modified = initial.replace()
    modified.attachments.append(
        model.Attachment(caption="The attachment", owner_group="owner")
    )
    assert initial != modified
    assert modified != initial


@given(sst.datasets())
@settings(max_examples=1)
def test_neq_mismatched_attachment(initial):
    initial.attachments = [
        (model.Attachment(caption="The attachment", owner_group="owner"))
    ]
    modified = initial.replace()
    modified.attachments[0] = model.Attachment(
        caption="Another attachment", owner_group="owner"
    )
    assert initial != modified
    assert modified != initial


@pytest.mark.parametrize(
    "field",
    (
        field
        for field in Dataset.fields(read_only=False)
        if field.name not in _UNGENERATABLE_FIELDS
    ),
    ids=lambda f: f.name,
)
@given(sst.datasets(), st.data())
@settings(max_examples=5)
def test_replace_replaces_single_writable_field(field, initial, data):
    val = data.draw(st.from_type(field.type))
    replaced = initial.replace(**{field.name: val})
    assert getattr(replaced, field.name) == val


@pytest.mark.parametrize(
    "field",
    (
        field
        for field in Dataset.fields(read_only=True)
        if field.name
        not in (
            "history",
            "lifecycle",
            "number_of_files",
            "number_of_files_archived",
            "size",
            "packed_size",
        )
    ),
    ids=lambda f: f.name,
)
@given(sst.datasets(), st.data())
@settings(max_examples=5)
def test_replace_replaces_single_read_only_field(field, initial, data):
    val = data.draw(st.from_type(field.type))
    replaced = initial.replace(_read_only={field.name: val})
    assert getattr(replaced, field.name) == val


@given(sst.datasets())
@settings(max_examples=5)
def test_replace_replaces_multiple_fields(initial):
    replaced = initial.replace(
        owner="a-new-owner",
        used_software=["software1"],
        _read_only={"created_by": "the-creator"},
    )
    assert replaced.owner == "a-new-owner"
    assert replaced.used_software == ["software1"]
    assert replaced.created_by == "the-creator"


@given(sst.datasets())
@settings(max_examples=5)
def test_replace_other_fields_are_copied(initial):
    replaced = initial.replace(
        investigator="inv@esti.gator",
        techniques=[model.Technique(pid="tech/abcd.01", name="magick")],
        type="raw",
        _read_only={"api_version": 666},
    )
    assert replaced.owner == initial.owner
    assert replaced.size == initial.size
    assert replaced.updated_at == initial.updated_at
    assert replaced.history == initial.history


@given(sst.datasets())
@settings(max_examples=1)
def test_replace_rejects_bad_arguments(initial):
    with pytest.raises(TypeError):
        initial.replace(this_is_not_a_valid="argument", owner="the-owner-of-it-all")


@given(sst.datasets())
@settings(max_examples=1)
def test_replace_does_not_change_files_no_input_files(initial):
    replaced = initial.replace(owner="a-new-owner")
    assert replaced.number_of_files == 0
    assert replaced.size == 0
    assert list(replaced.files) == []


@given(sst.datasets())
@settings(max_examples=1)
def test_replace_does_not_change_files_with_input_files(initial):
    file = File.from_download_model(
        local_path=None,
        model=model.DownloadDataFile(path="path", size=6163, time=datetime.now()),
    )
    initial.add_files(file)
    replaced = initial.replace(owner="a-new-owner")
    assert replaced.number_of_files == 1
    assert replaced.size == 6163
    assert list(replaced.files) == list(initial.files)


@given(sst.datasets())
@settings(max_examples=1)
def test_replace_preserves_meta(initial):
    initial.meta["key"] = "val"
    replaced = initial.replace(owner="a-new-owner")
    assert replaced.meta == {"key": "val"}


@given(sst.datasets())
@settings(max_examples=1)
def test_replace_meta(initial):
    initial.meta["key"] = {"value": 2, "unit": "m"}
    initial.meta["old-key"] = "old-val"
    replaced = initial.replace(
        owner="a-new-owner",
        meta={"key": {"value": 3, "unit": "m"}, "new-key": "new-val"},
    )
    assert replaced.meta == {"key": {"value": 3, "unit": "m"}, "new-key": "new-val"}


@given(sst.datasets())
@settings(max_examples=1)
def test_replace_remove_meta(initial):
    initial.meta["key"] = {"value": 2, "unit": "m"}
    initial.meta["old-key"] = "old-val"
    replaced = initial.replace(owner="a-new-owner", meta=None)
    assert replaced.meta == {}


@pytest.mark.parametrize(
    "attachments",
    (None, [], [model.Attachment(caption="Attachment 1", owner_group="owner")]),
)
@given(initial=sst.datasets())
@settings(max_examples=1)
def test_replace_preserves_attachments(initial, attachments):
    initial.attachments = attachments
    replaced = initial.replace(owner="a-new-owner")
    assert replaced.attachments == attachments


@pytest.mark.parametrize(
    "attachments",
    (None, [], [model.Attachment(caption="Attachment 1", owner_group="owner")]),
)
@pytest.mark.parametrize(
    "target_attachments",
    (None, [], [model.Attachment(caption="Attachment 2", owner_group="owner")]),
)
@given(initial=sst.datasets())
@settings(max_examples=1)
def test_replace_attachments(initial, attachments, target_attachments):
    replaced = initial.replace(attachments=target_attachments)
    assert replaced.attachments == target_attachments


@given(sst.datasets())
@settings(max_examples=5)
def test_as_new(initial):
    new = initial.as_new()
    assert new.created_at is None
    assert new.created_by is None
    assert new.updated_at is None
    assert new.updated_by is None
    assert new.history is None
    assert new.lifecycle is None
    assert abs(new.creation_time - datetime.now(tz=timezone.utc)) < timedelta(seconds=1)

    assert new.number_of_files == initial.number_of_files
    assert new.size == initial.size
    assert new.name == initial.name
    assert new.input_datasets == initial.input_datasets
    assert new.data_format == initial.data_format


@given(sst.datasets(pid=PID(pid="some-id")))
@settings(max_examples=5)
def test_derive_default(initial):
    derived = initial.derive()
    assert derived.type == "derived"
    assert derived.input_datasets == [initial.pid]
    assert derived.lifecycle is None
    assert derived.history is None

    assert derived.investigator == initial.investigator
    assert derived.owner == initial.owner
    assert derived.orcid_of_owner == initial.orcid_of_owner
    assert derived.owner_email == initial.owner_email
    assert derived.contact_email == initial.contact_email
    assert derived.techniques == initial.techniques

    assert derived.name is None
    assert derived.used_software is None
    assert derived.number_of_files == 0
    assert derived.number_of_files_archived == 0


@given(sst.datasets(pid=PID(pid="some-id")))
@settings(max_examples=5)
def test_derive_set_keep(initial):
    derived = initial.derive(keep=("name", "used_software"))
    assert derived.type == "derived"
    assert derived.input_datasets == [initial.pid]
    assert derived.lifecycle is None
    assert derived.history is None

    assert derived.name == initial.name
    assert derived.used_software == initial.used_software

    assert derived.number_of_files == 0
    assert derived.number_of_files_archived == 0


@given(sst.datasets(pid=PID(pid="some-id")))
@settings(max_examples=5)
def test_derive_keep_nothing(initial):
    derived = initial.derive(keep=())
    assert derived.type == "derived"
    assert derived.input_datasets == [initial.pid]
    assert derived.lifecycle is None
    assert derived.history is None

    assert derived.investigator is None
    assert derived.owner is None
    assert derived.name is None
    assert derived.used_software is None
    assert derived.number_of_files == 0
    assert derived.number_of_files_archived == 0


@given(sst.datasets(pid=None))
@settings(max_examples=5)
def test_derive_requires_pid(initial):
    with pytest.raises(ValueError):
        initial.derive()


@pytest.mark.parametrize(
    "attachments",
    (None, [], [model.Attachment(caption="Attachment 1", owner_group="owner")]),
)
@given(initial=sst.datasets(pid=PID(pid="some-id")))
@settings(max_examples=1)
def test_derive_removes_attachments(initial, attachments):
    initial.attachments = attachments
    derived = initial.derive()
    assert derived.attachments == []


def invalid_field_example(my_type):
    if my_type == DatasetType.DERIVED:
        return "data_format", "sth_not_None"
    elif my_type == DatasetType.RAW:
        return "job_log_data", "sth_not_None"
    else:
        raise ValueError(my_type, " is not valid DatasetType.")


@given(initial=sst.datasets(for_upload=True))
@settings(max_examples=10)
def test_dataset_dict_like_keys_per_type(initial: Dataset) -> None:
    my_names = set(
        field.name for field in Dataset._FIELD_SPEC if field.used_by(initial.type)
    )
    assert set(initial.keys()) == my_names


@given(initial=sst.datasets(for_upload=True))
@settings(max_examples=10)
def test_dataset_dict_like_keys_including_invalid_field(initial):
    invalid_name, invalid_value = invalid_field_example(initial.type)

    my_names = set(
        field.name for field in Dataset._FIELD_SPEC if field.used_by(initial.type)
    )
    assert invalid_name not in my_names
    my_names.add(invalid_name)

    setattr(initial, invalid_name, invalid_value)

    assert set(initial.keys()) == my_names


@given(initial=sst.datasets(for_upload=True))
@settings(max_examples=10)
def test_dataset_dict_like_values(initial: Dataset) -> None:
    for key, value in zip(initial.keys(), initial.values()):
        assert value == getattr(initial, key)


@given(initial=sst.datasets(for_upload=True))
@settings(max_examples=10)
def test_dataset_dict_like_values_with_invalid_field(initial: Dataset) -> None:
    setattr(initial, *invalid_field_example(initial.type))
    for key, value in zip(initial.keys(), initial.values()):
        assert value == getattr(initial, key)


@given(initial=sst.datasets(for_upload=True))
@settings(max_examples=10)
def test_dataset_dict_like_items_with_invalid_field(initial: Dataset) -> None:
    setattr(initial, *invalid_field_example(initial.type))
    for key, value in initial.items():
        assert value == getattr(initial, key)


@given(initial=sst.datasets(for_upload=True))
@settings(max_examples=10)
def test_dataset_dict_like_getitem(initial):
    assert initial["type"] == initial.type


@pytest.mark.parametrize(
    ("is_attr", "wrong_field"), ((True, "size"), (False, "OBVIOUSLYWRONGNAME"))
)
@given(initial=sst.datasets(for_upload=True))
@settings(max_examples=10)
def test_dataset_dict_like_getitem_wrong_field_raises(initial, is_attr, wrong_field):
    # 'size' should be included in the field later.
    # It is now excluded because it is ``manual`` field. See issue#151.
    assert hasattr(initial, wrong_field) == is_attr
    with pytest.raises(KeyError, match=f"{wrong_field} is not a valid field name."):
        initial[wrong_field]


@given(initial=sst.datasets(for_upload=True))
@settings(max_examples=10)
def test_dataset_dict_like_setitem(initial: Dataset) -> None:
    import uuid

    sample_comment = uuid.uuid4().hex
    assert initial["comment"] != sample_comment
    initial["comment"] = sample_comment
    assert initial["comment"] == sample_comment


@given(initial=sst.datasets(for_upload=True))
@settings(max_examples=10)
def test_dataset_dict_like_setitem_invalid_field(initial: Dataset) -> None:
    # ``__setitem__`` doesn't check if the item is invalid for the current type or not.
    invalid_field, invalid_value = invalid_field_example(initial.type)
    assert initial[invalid_field] is None
    initial[invalid_field] = invalid_value
    assert initial[invalid_field] == invalid_value


@pytest.mark.parametrize(
    ("is_attr", "wrong_field", "wrong_value"),
    ((True, "size", 10), (False, "OBVIOUSLYWRONGNAME", "OBVIOUSLYWRONGVALUE")),
)
@given(initial=sst.datasets(for_upload=True))
@settings(max_examples=10)
def test_dataset_dict_like_setitem_wrong_field_raises(
    initial, is_attr, wrong_field, wrong_value
):
    # ``manual`` fields such as ``size`` should raise with ``__setitem__``.
    # However, it may need more specific error message.
    assert hasattr(initial, wrong_field) == is_attr
    with pytest.raises(KeyError, match=f"{wrong_field} is not a valid field name."):
        initial[wrong_field] = wrong_value



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
# mypy: disable-error-code="arg-type, index"

import pydantic
import pytest
from dateutil.parser import parse as parse_date

from scitacean import PID, Client, RemotePath, ScicatCommError
from scitacean.client import ScicatClient
from scitacean.model import (
    DatasetType,
    UploadDerivedDataset,
)
from scitacean.testing.backend.seed import (
    INITIAL_DATASETS,
    INITIAL_ORIG_DATABLOCKS,
)


@pytest.fixture()
def scicat_client(client: Client) -> ScicatClient:
    return client.scicat


@pytest.fixture()
def derived_dataset(scicat_access):
    return UploadDerivedDataset(
        contactEmail="black.foess@dom.koelle",
        creationTime=parse_date("1995-11-11T11:11:11.000Z"),
        owner="bfoess",
        investigator="b.foess@dom.koelle",
        sourceFolder="/dom/platt",
        type=DatasetType.DERIVED,
        inputDatasets=[],
        usedSoftware=[],
        ownerGroup=scicat_access.user.group,
        accessGroups=["koelle"],
        numberOfFilesArchived=0,
    )


@pytest.mark.parametrize("key", ("raw", "derived"))
def test_get_dataset_model(scicat_client, key):
    dset = INITIAL_DATASETS[key]
    downloaded = scicat_client.get_dataset_model(dset.pid)
    # The backend may update the dataset after upload.
    # We cannot easily predict when that happens.
    downloaded.updatedAt = dset.updatedAt
    assert downloaded == dset


def test_get_dataset_model_bad_id(scicat_client):
    with pytest.raises(ScicatCommError):
        scicat_client.get_dataset_model(PID(pid="bad-pid"))


def test_create_dataset_model(scicat_client, derived_dataset):
    finalized = scicat_client.create_dataset_model(derived_dataset)
    downloaded = scicat_client.get_dataset_model(finalized.pid)
    for key, expected in finalized:
        # The database populates a number of fields that are None in dset.
        # But we don't want to test those here as we don't want to test the database.
        if expected is not None:
            assert expected == dict(downloaded)[key], f"key = {key}"


def test_validate_dataset_model(real_client, require_scicat_backend, derived_dataset):
    real_client.scicat.validate_dataset_model(derived_dataset)
    derived_dataset.contactEmail = "NotAnEmail"
    with pytest.raises(ValueError):
        real_client.scicat.validate_dataset_model(derived_dataset)


def test_get_dataset(client):
    dset = INITIAL_DATASETS["raw"]
    dblock = INITIAL_ORIG_DATABLOCKS["raw"][0]
    downloaded = client.get_dataset(dset.pid)

    assert downloaded.source_folder == dset.sourceFolder
    assert downloaded.creation_time == dset.creationTime
    assert downloaded.access_groups == dset.accessGroups
    assert downloaded.meta["temperature"] == dset.scientificMetadata["temperature"]
    assert downloaded.meta["data_type"] == dset.scientificMetadata["data_type"]

    for dset_file, expected_file in zip(downloaded.files, dblock.dataFileList):
        assert dset_file.local_path is None
        assert dset_file.size == expected_file.size
        assert dset_file.creation_time == expected_file.time


def test_can_get_public_dataset_without_login(require_scicat_backend, scicat_access):
    client = Client.without_login(url=scicat_access.url)

    dset = INITIAL_DATASETS["public"]
    dblock = INITIAL_ORIG_DATABLOCKS["public"][0]
    downloaded = client.get_dataset(dset.pid)

    assert downloaded.source_folder == dset.sourceFolder
    assert downloaded.creation_time == dset.creationTime
    assert downloaded.access_groups == dset.accessGroups

    for dset_file, expected_file in zip(downloaded.files, dblock.dataFileList):
        assert dset_file.local_path is None
        assert dset_file.size == expected_file.size
        assert dset_file.creation_time == expected_file.time


def test_cannot_upload_without_login(
    require_scicat_backend, derived_dataset, scicat_access
):
    client = Client.without_login(url=scicat_access.url)
    with pytest.raises(ScicatCommError):  # TODO test return code 403
        client.scicat.create_dataset_model(derived_dataset)


def test_get_broken_dataset(client):
    dset = INITIAL_DATASETS["partially-broken"]
    downloaded = client.get_dataset(dset.pid)
    assert downloaded.type == DatasetType.DERIVED
    # Intact field; was properly converted to RemotePath
    assert isinstance(downloaded.source_folder, RemotePath)
    assert downloaded.source_folder == "/remote/source"

    # Broken fields loaded
    assert isinstance(downloaded.orcid_of_owner, str)
    assert downloaded.orcid_of_owner == "00-11-22-33"

    # Ignored broken fields
    assert isinstance(downloaded.number_of_files, int)
    assert downloaded.number_of_files == 0
    assert isinstance(downloaded.size, int)
    assert downloaded.size == 0


def test_get_broken_dataset_strict_validation(real_client, require_scicat_backend):
    dset = INITIAL_DATASETS["partially-broken"]
    with pytest.raises(pydantic.ValidationError):
        real_client.get_dataset(dset.pid, strict_validation=True)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scitacean contributors (https://github.com/SciCatProject/scitacean)
# mypy: disable-error-code="no-untyped-def, return-value, arg-type, union-attr"

import dataclasses
import tempfile
from contextlib import contextmanager
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Callable, Iterator, Optional

import fabric
import paramiko
import pytest
from fabric import Connection

from scitacean import Dataset, File, FileUploadError, RemotePath
from scitacean.testing.client import FakeClient
from scitacean.testing.ssh import IgnorePolicy, skip_if_not_ssh
from scitacean.transfer.ssh import (
    SSHDownloadConnection,
    SSHFileTransfer,
    SSHUploadConnection,
)


@pytest.fixture(scope="session", autouse=True)
def server(request, ssh_fileserver):
    skip_if_not_ssh(request)


def test_download_one_file(ssh_access, ssh_connect_with_username_password, tmp_path):
    ssh = SSHFileTransfer(host=ssh_access.host, port=ssh_access.port)
    with ssh.connect_for_download(connect=ssh_connect_with_username_password) as con:
        con.download_files(
            remote=[RemotePath("/data/seed/text.txt")], local=[tmp_path / "text.txt"]
        )
    with open(tmp_path / "text.txt", "r") as f:
        assert f.read() == "This is some text for testing.\n"


def test_download_two_files(ssh_access, ssh_connect_with_username_password, tmp_path):
    ssh = SSHFileTransfer(host=ssh_access.host, port=ssh_access.port)
    with ssh.connect_for_download(connect=ssh_connect_with_username_password) as con:
        con.download_files(
            remote=[
                RemotePath("/data/seed/table.csv"),
                RemotePath("/data/seed/text.txt"),
            ],
            local=[tmp_path / "local-table.csv", tmp_path / "text.txt"],
        )
    with open(tmp_path / "local-table.csv", "r") as f:
        assert f.read() == "7,2\n5,2\n"
    with open(tmp_path / "text.txt", "r") as f:
        assert f.read() == "This is some text for testing.\n"


def test_upload_one_file_source_folder_in_dataset(
    ssh_access, ssh_connect_with_username_password, tmp_path, ssh_data_dir
):
    ds = Dataset(type="raw", source_folder=RemotePath("/data/upload"))
    with open(tmp_path / "file0.txt", "w") as f:
        f.write("File to test upload123")

    ssh = SSHFileTransfer(host=ssh_access.host, port=ssh_access.port)
    with ssh.connect_for_upload(
        dataset=ds, connect=ssh_connect_with_username_password
    ) as con:
        assert con.source_folder == RemotePath("/data/upload")
        con.upload_files(
            File.from_local(path=tmp_path / "file0.txt", remote_path="upload_0.txt")
        )

    with open(ssh_data_dir / "upload" / "upload_0.txt", "r") as f:
        assert f.read() == "File to test upload123"


def test_upload_one_file_source_folder_in_transfer(
    ssh_access, ssh_connect_with_username_password, tmp_path, ssh_data_dir
):
    ds = Dataset(type="raw", owner="librarian")
    with open(tmp_path / "file1.txt", "w") as f:
        f.write("File no. 2")

    ssh = SSHFileTransfer(
        host=ssh_access.host,
        port=ssh_access.port,
        source_folder="/data/upload/{owner}",
    )
    with ssh.connect_for_upload(
        dataset=ds, connect=ssh_connect_with_username_password
    ) as con:
        assert con.source_folder == RemotePath("/data/upload/librarian")
        con.upload_files(
            File.from_local(
                path=tmp_path / "file1.txt", remote_path=RemotePath("upload_1.txt")
            )
        )

    with open(ssh_data_dir / "upload" / "librarian" / "upload_1.txt", "r") as f:
        assert f.read() == "File no. 2"


def test_upload_two_files(
    ssh_access, ssh_connect_with_username_password, tmp_path, ssh_data_dir
):
    ds = Dataset(type="raw", source_folder=RemotePath("/data/upload2"))
    with open(tmp_path / "file2.1.md", "w") as f:
        f.write("First part of file 2")
    with open(tmp_path / "file2.2.md", "w") as f:
        f.write("Second part of file 2")

    ssh = SSHFileTransfer(host=ssh_access.host, port=ssh_access.port)
    with ssh.connect_for_upload(
        dataset=ds, connect=ssh_connect_with_username_password
    ) as con:
        assert con.source_folder == RemotePath("/data/upload2")
        con.upload_files(
            File.from_local(path=tmp_path / "file2.1.md", base_path=tmp_path),
            File.from_local(path=tmp_path / "file2.2.md", base_path=tmp_path),
        )

    with open(ssh_data_dir / "upload2" / "file2.1.md", "r") as f:
        assert f.read() == "First part of file 2"
    with open(ssh_data_dir / "upload2" / "file2.2.md", "r") as f:
        assert f.read() == "Second part of file 2"


def test_revert_all_uploaded_files_single(
    ssh_access, ssh_connect_with_username_password, tmp_path, ssh_data_dir
):
    ds = Dataset(type="raw", source_folder=RemotePath("/data/revert-all-test-1"))
    with open(tmp_path / "file3.txt", "w") as f:
        f.write("File that should get reverted")

    ssh = SSHFileTransfer(host=ssh_access.host, port=ssh_access.port)
    with ssh.connect_for_upload(
        dataset=ds, connect=ssh_connect_with_username_password
    ) as con:
        file = File.from_local(path=tmp_path / "file3.txt", base_path=tmp_path)
        con.upload_files(file)
        con.revert_upload(file)

    assert "revert-all-test-1" not in (p.name for p in ssh_data_dir.iterdir())


def test_revert_all_uploaded_files_two(
    ssh_access, ssh_connect_with_username_password, tmp_path, ssh_data_dir
):
    ds = Dataset(type="raw", source_folder=RemotePath("/data/revert-all-test-2"))
    with open(tmp_path / "file3.1.txt", "w") as f:
        f.write("File that should get reverted 1")
    with open(tmp_path / "file3.2.txt", "w") as f:
        f.write("File that should get reverted 2")

    ssh = SSHFileTransfer(host=ssh_access.host, port=ssh_access.port)
    with ssh.connect_for_upload(
        dataset=ds, connect=ssh_connect_with_username_password
    ) as con:
        file1 = File.from_local(path=tmp_path / "file3.1.txt", base_path=tmp_path)
        file2 = File.from_local(path=tmp_path / "file3.2.txt", base_path=tmp_path)
        con.upload_files(file1, file2)
        con.revert_upload(file1, file2)

    assert "revert-all-test-2" not in (p.name for p in ssh_data_dir.iterdir())


def test_revert_one_uploaded_file(
    ssh_access, ssh_connect_with_username_password, tmp_path, ssh_data_dir
):
    ds = Dataset(type="raw", source_folder=RemotePath("/data/revert-test"))
    with open(tmp_path / "file4.txt", "w") as f:
        f.write("File that should get reverted")
    with open(tmp_path / "file5.txt", "w") as f:
        f.write("File that should be kept")

    ssh = SSHFileTransfer(host=ssh_access.host, port=ssh_access.port)
    with ssh.connect_for_upload(
        dataset=ds, connect=ssh_connect_with_username_password
    ) as con:
        file4 = File.from_local(path=tmp_path / "file4.txt", base_path=tmp_path)
        file5 = File.from_local(path=tmp_path / "file5.txt", base_path=tmp_path)
        con.upload_files(file4, file5)
        con.revert_upload(file4)

    assert "file4.txt" not in (p.name for p in (ssh_data_dir / "revert-test").iterdir())
    with open(ssh_data_dir / "revert-test" / "file5.txt", "r") as f:
        assert f.read() == "File that should be kept"


def test_stat_uploaded_file(
    ssh_access, ssh_connect_with_username_password, tmp_path, ssh_data_dir
):
    ds = Dataset(type="raw", source_folder=RemotePath("/data/upload6"))
    with open(tmp_path / "file6.txt", "w") as f:
        f.write("File to test upload no 6")

    ssh = SSHFileTransfer(host=ssh_access.host, port=ssh_access.port)
    with ssh.connect_for_upload(
        dataset=ds, connect=ssh_connect_with_username_password
    ) as con:
        [uploaded] = con.upload_files(
            File.from_local(path=tmp_path / "file6.txt", remote_path="upload_6.txt")
        )

    st = (ssh_data_dir / "upload6" / "upload_6.txt").stat()
    assert uploaded.size == st.st_size

    # Set in docker-compose
    assert uploaded.remote_uid == "1000"
    assert uploaded.remote_gid == "1000"

    uploaded = dataclasses.replace(uploaded, local_path=None)
    assert datetime.now(tz=timezone.utc) - uploaded.creation_time < timedelta(seconds=5)


class CorruptingSFTP(paramiko.SFTPClient):
    """Appends bytes to uploaded files to simulate a broken transfer."""

    def put(self, localpath, remotepath, callback=None, confirm=True):
        with open(localpath, "r") as f:
            content = f.read()
        with tempfile.TemporaryDirectory() as tempdir:
            corrupted_path = Path(tempdir) / "corrupted"
            with open(corrupted_path, "w") as f:
                f.write(content + "\nevil bytes")
            super().put(str(corrupted_path), remotepath, callback, confirm)


class CorruptingTransfer(paramiko.Transport):
    """Uses CorruptingSFTP to simulate a broken connection."""

    def open_sftp_client(self) -> paramiko.SFTPClient:
        return CorruptingSFTP.from_transport(self)


@pytest.fixture()
def ssh_corrupting_connect(ssh_access, ssh_connection_config):
    def connect(host: str, port: int, **kwargs):
        if kwargs:
            raise ValueError(
                "connect_with_username_password must only be"
                f" used without extra arguments. Got {kwargs=}"
            )
        connection = fabric.Connection(
            host=host,
            port=port,
            user=ssh_access.user.username,
            config=ssh_connection_config,
            connect_kwargs={
                "password": ssh_access.user.password,
                "transport_factory": CorruptingTransfer,
                **ssh_connection_config.connect_kwargs,
            },
        )
        connection.client.set_missing_host_key_policy(IgnorePolicy())
        return connection

    return connect


def test_upload_file_detects_checksum_mismatch(
    ssh_access, ssh_corrupting_connect, tmp_path, ssh_data_dir
):
    ds = Dataset(
        type="raw",
        source_folder=RemotePath("/data/upload7"),
        checksum_algorithm="blake2b",
    )
    with open(tmp_path / "file7.txt", "w") as f:
        f.write("File to test upload no 7")

    ssh = SSHFileTransfer(host=ssh_access.host, port=ssh_access.port)
    with ssh.connect_for_upload(dataset=ds, connect=ssh_corrupting_connect) as con:
        with pytest.raises(FileUploadError):
            con.upload_files(
                dataclasses.replace(
                    File.from_local(
                        path=tmp_path / "file7.txt",
                        remote_path=RemotePath("upload_7.txt"),
                    ),
                    checksum_algorithm="blake2b",
                )
            )

    assert "upload7" not in (p.name for p in ssh_data_dir.iterdir())


class RaisingSFTP(paramiko.SFTPClient):
    def put(self, localpath, remotepath, callback=None, confirm=True):
        raise RuntimeError("Upload disabled")


class RaisingTransfer(paramiko.Transport):
    def open_sftp_client(self) -> paramiko.SFTPClient:
        return RaisingSFTP.from_transport(self)


@pytest.fixture()
def ssh_raising_connect(ssh_access, ssh_connection_config):
    def connect(host: str, port: int, **kwargs):
        if kwargs:
            raise ValueError(
                "connect_with_username_password must only be"
                f" used without extra arguments. Got {kwargs=}"
            )
        connection = fabric.Connection(
            host=host,
            port=port,
            user=ssh_access.user.username,
            config=ssh_connection_config,
            connect_kwargs={
                "password": ssh_access.user.password,
                "transport_factory": RaisingTransfer,
                **ssh_connection_config.connect_kwargs,
            },
        )
        connection.client.set_missing_host_key_policy(IgnorePolicy())
        return connection

    return connect


def test_upload_file_reverts_if_upload_fails(
    ssh_access, ssh_raising_connect, tmp_path, ssh_data_dir
):
    ds = Dataset(type="raw", source_folder=RemotePath("/data/upload8"))
    with open(tmp_path / "file8.txt", "w") as f:
        f.write("File to test upload no 8")

    ssh = SSHFileTransfer(host=ssh_access.host, port=ssh_access.port)
    with ssh.connect_for_upload(dataset=ds, connect=ssh_raising_connect) as con:
        with pytest.raises(RuntimeError):
            con.upload_files(
                File.from_local(
                    path=tmp_path / "file8.txt",
                    remote_path=RemotePath("upload_8.txt"),
                )
            )

    assert "upload8" not in (p.name for p in ssh_data_dir.iterdir())


class SSHTestFileTransfer(SSHFileTransfer):
    def __init__(self, connect, **kwargs):
        super().__init__(**kwargs)
        self.connect = connect

    @contextmanager
    def connect_for_download(
        self, connect: Optional[Callable[..., Connection]] = None
    ) -> Iterator[SSHDownloadConnection]:
        connect = connect if connect is not None else self.connect
        with super().connect_for_download(connect=connect) as connection:
            yield connection

    @contextmanager
    def connect_for_upload(
        self, dataset: Dataset, connect: Optional[Callable[..., Connection]] = None
    ) -> Iterator[SSHUploadConnection]:
        connect = connect if connect is not None else self.connect
        with super().connect_for_upload(dataset=dataset, connect=connect) as connection:
            yield connection


# This test is referenced in the docs.
def test_client_with_ssh(
    require_ssh_fileserver,
    ssh_access,
    ssh_connect_with_username_password,
    ssh_data_dir,
    tmp_path,
):
    tmp_path.joinpath("file1.txt").write_text("File contents")

    client = FakeClient.without_login(
        url="",
        file_transfer=SSHTestFileTransfer(
            connect=ssh_connect_with_username_password,
            host=ssh_access.host,
            port=ssh_access.port,
        ),
    )
    ds = Dataset(
        access_groups=["group1"],
        contact_email="p.stibbons@uu.am",
        creation_location="UU",
        creation_time=datetime(2023, 6, 23, 10, 0, 0),
        owner="PonderStibbons",
        owner_group="uu",
        principal_investigator="MustrumRidcully",
        source_folder="/data",
        type="raw",
    )
    ds.add_local_files(tmp_path / "file1.txt", base_path=tmp_path)
    finalized = client.upload_new_dataset_now(ds)

    downloaded = client.get_dataset(finalized.pid)
    downloaded = client.download_files(downloaded, target=tmp_path / "download")

    assert ssh_data_dir.joinpath("file1.txt").read_text() == "File contents"
    assert downloaded.files[0].local_path.read_text() == "File contents"



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
# mypy: disable-error-code="no-untyped-def"
"""Pytest fixtures to manage and access a local SFTP server."""

import logging
from pathlib import Path
from typing import Callable, Generator, Optional

import pytest
from paramiko import SFTPClient, SSHClient

from ..._internal import docker
from .._pytest_helpers import init_work_dir, root_tmp_dir
from ._pytest_helpers import sftp_enabled, skip_if_not_sftp
from ._sftp import (
    IgnorePolicy,
    SFTPAccess,
    configure,
    local_access,
    wait_until_sftp_server_is_live,
)


@pytest.fixture(scope="session")
def sftp_access(request: pytest.FixtureRequest) -> SFTPAccess:
    """Fixture that returns SFTP access parameters.

    Returns
    -------
    :
        A URL and user to connect to the testing SFTP server.
        The user has access to all initial files registered in the
        database and permissions to create new files.
    """
    skip_if_not_sftp(request)
    return local_access()


@pytest.fixture(scope="session")
def sftp_base_dir(
    request: pytest.FixtureRequest, tmp_path_factory: pytest.TempPathFactory
) -> Optional[Path]:
    """Fixture that returns the base working directory for the SFTP server setup.

    Returns
    -------
    :
        A path to a directory on the host machine.
        The directory gets populated by the
        :func:`scitacean.testing.sftp.fixtures.sftp_fileserver` fixture.
        It contains the docker configuration and volumes.

        Returns ``None`` if SFTP tests are disabled
    """
    if not sftp_enabled(request):
        return None
    return root_tmp_dir(request, tmp_path_factory) / "scitacean-sftp"


@pytest.fixture(scope="session")
def sftp_data_dir(sftp_base_dir: Optional[Path]) -> Optional[Path]:
    """Fixture that returns the data directory for the SFTP server setup.

    Returns
    -------
    :
        A path to a directory on the host machine.
        The directory is mounted as ``/data`` on the server.

        Returns ``None`` if SFTP tests are disabled
    """
    if sftp_base_dir is None:
        return None
    return sftp_base_dir / "data"


@pytest.fixture()
def require_sftp_fileserver(request, sftp_fileserver) -> None:
    """Fixture to declare that a test needs a local SFTP server.

    Like :func:`scitacean.testing.sftp.sftp_fileserver`
    but this skips the test if SFTP tests are disabled.
    """
    skip_if_not_sftp(request)


@pytest.fixture(scope="session")
def sftp_fileserver(
    request: pytest.FixtureRequest,
    sftp_access: SFTPAccess,
    sftp_base_dir: Optional[Path],
    sftp_data_dir: Optional[Path],
    sftp_connect_with_username_password,
) -> Generator[bool, None, None]:
    """Fixture to declare that a test needs a local SFTP server.

    If SFTP tests are enabled, this fixture configures and starts an SFTP server
    in a docker container the first time a test requests it.
    The server and container will be stopped and removed at the end of the test session.

    Does nothing if the SFTP tests are disabled.

    Returns
    -------
    :
        True if SFTP tests are enabled and False otherwise.
    """
    if sftp_base_dir is None:
        yield False
        return

    target_dir, counter = init_work_dir(request, sftp_base_dir, name=None)

    try:
        with counter.increment() as count:
            if count == 1:
                _sftp_docker_up(target_dir, sftp_access)
            elif not _sftp_server_is_running():
                raise RuntimeError("Expected SFTP server to be running")
        yield True
    finally:
        with counter.decrement() as count:
            if count == 0:
                _sftp_docker_down(target_dir)


@pytest.fixture(scope="session")
def sftp_connect_with_username_password(
    sftp_access: SFTPAccess,
) -> Callable[[str, int], SFTPClient]:
    """Fixture that returns a function to connect to the testing SFTP server.

    Uses username+password from the test config.

    Returns
    -------
    :
        A function to pass as the ``connect`` argument when constructing a
        :class:`scitacean.transfer.sftp.SFTPFileTransfer`.

    Examples
    --------
    Explicitly connect to the test server:

    .. code-block:: python

        def test_sftp(sftp_access,
                      sftp_connect_with_username_password,
                      sftp_fileserver):
            sftp = SFTPFileTransfer(host=sftp_access.host,
                                    port=sftp_access.port,
                                    connect=sftp_connect_with_username_password)
            with sftp.connect_for_download() as connection:
                # use connection
    """

    def connect(host: str, port: int) -> SFTPClient:
        client = SSHClient()
        client.set_missing_host_key_policy(IgnorePolicy())
        client.connect(
            hostname=host,
            port=port,
            username=sftp_access.user.username,
            password=sftp_access.user.password,
            allow_agent=False,
            look_for_keys=False,
        )
        return client.open_sftp()

    return connect


def _sftp_docker_up(target_dir: Path, sftp_access: SFTPAccess) -> None:
    if _sftp_server_is_running():
        raise RuntimeError("SFTP docker container is already running")
    docker_compose_file = target_dir / "docker-compose.yaml"
    log = logging.getLogger("scitacean.testing")
    log.info("Starting docker container with SFTP server from %s", docker_compose_file)
    configure(target_dir)
    docker.docker_compose_up(docker_compose_file)
    log.info("Waiting for SFTP docker to become accessible")
    wait_until_sftp_server_is_live(sftp_access=sftp_access, max_time=60, n_tries=40)
    log.info("Successfully connected to SFTP server")
    # Give the user write access.
    docker.docker_compose_run(
        docker_compose_file, "scitacean-test-sftp-server", "chown", "1000:1000", "/data"
    )


def _sftp_docker_down(target_dir: Path) -> None:
    # Check if container is running because the fixture can call this function
    # if there was an exception in _sftp_docker_up.
    # In that case, there is nothing to tear down.
    if _sftp_server_is_running():
        docker_compose_file = target_dir / "docker-compose.yaml"
        log = logging.getLogger("scitacean.testing")
        log.info(
            "Stopping docker container with SFTP server from %s", docker_compose_file
        )
        docker.docker_compose_down(docker_compose_file)


def _sftp_server_is_running() -> bool:
    return docker.container_is_running("scitacean-test-sftp")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

from contextlib import contextmanager
from typing import cast

import pytest
from dateutil.parser import parse as parse_date

from scitacean import (
    PID,
    Attachment,
    Client,
    Dataset,
    DatasetType,
    File,
    ScicatCommError,
    Thumbnail,
)
from scitacean.testing.client import FakeClient
from scitacean.testing.transfer import FakeFileTransfer

from .common.files import make_file

# TODO source_folder changed / populated in file after upload


def get_file_transfer(client: Client) -> FakeFileTransfer:
    return client.file_transfer  # type: ignore[return-value]


@pytest.fixture
def dataset():
    return Dataset(
        access_groups=["group1", "2nd_group"],
        investigator="ridcully@uu.am",
        contact_email="p.stibbons@uu.am",
        source_folder="/hex/source123",
        creation_time=parse_date("2011-08-24T12:34:56Z"),
        input_datasets=[],
        name="Data A38",
        owner="PonderStibbons",
        owner_group="uu",
        used_software=["EasyScience"],
        meta={
            "data_type": "event data",
            "temperature": {"value": "123", "unit": "K"},
            "weight": {"value": "42", "unit": "mg"},
        },
        type=DatasetType.DERIVED,
    )


@pytest.fixture
def dataset_with_files(dataset, fs):
    make_file(fs, path="file.nxs", contents=b"contents of file.nxs")
    make_file(fs, path="the_log_file.log", contents=b"this is a log file")
    dataset.add_local_files("file.nxs", "the_log_file.log")
    return dataset


@pytest.fixture
def attachments():
    return [
        Attachment(
            caption="Attachment no 1",
            owner_group="uu",
            thumbnail=Thumbnail(mime="png/jpeg", data=b"840109725761"),
        ),
        Attachment(
            caption="Second attachment",
            owner_group="uu",
            thumbnail=Thumbnail(mime=None, data=b"5189762957"),
        ),
    ]


@pytest.fixture
def client(fs, scicat_access):
    return FakeClient.from_credentials(
        url="",
        **scicat_access.user.credentials,
        file_transfer=FakeFileTransfer(fs=fs, files={}, reverted={}),
    )


def test_upload_returns_updated_dataset_no_attachments(client, dataset_with_files):
    finalized = client.upload_new_dataset_now(dataset_with_files)
    expected = client.get_dataset(finalized.pid, attachments=True).replace(
        # The backend may update the dataset after upload
        _read_only={
            "updated_at": finalized.updated_at,
            "updated_by": finalized.updated_by,
        }
    )
    assert finalized == expected


def test_upload_returns_updated_dataset_with_attachments(
    client, dataset_with_files, attachments
):
    dataset_with_files.attachments = attachments
    finalized = client.upload_new_dataset_now(dataset_with_files)
    expected = client.get_dataset(finalized.pid, attachments=True).replace(
        # The backend may update the dataset after upload
        _read_only={
            "updated_at": finalized.updated_at,
            "updated_by": finalized.updated_by,
        }
    )
    assert finalized == expected


def test_upload_without_files_creates_dataset(client, dataset):
    finalized = client.upload_new_dataset_now(dataset)
    expected = client.get_dataset(finalized.pid, attachments=True).replace(
        # The backend may update the dataset after upload
        _read_only={
            "updated_at": finalized.updated_at,
            "updated_by": finalized.updated_by,
        }
    )
    assert finalized == expected
    with pytest.raises(ScicatCommError):
        client.scicat.get_orig_datablocks(finalized.pid)


def test_upload_without_files_does_not_need_file_transfer(dataset):
    client = FakeClient()
    finalized = client.upload_new_dataset_now(dataset)
    pid = cast(PID, finalized.pid)
    expected = client.get_dataset(pid, attachments=True).replace(
        # The backend may update the dataset after upload
        _read_only={
            "updated_at": finalized.updated_at,
            "updated_by": finalized.updated_by,
        }
    )
    assert finalized == expected
    with pytest.raises(ScicatCommError):
        client.scicat.get_orig_datablocks(pid)


def test_upload_without_files_does_not_need_revert_files(dataset):
    client = FakeClient(
        disable={"create_dataset_model": ScicatCommError("Ingestion failed")}
    )
    with pytest.raises(ScicatCommError):
        # Does not raise from attempting to access the file transfer.
        client.upload_new_dataset_now(dataset)


def test_upload_with_only_remote_files_does_not_need_file_transfer(dataset):
    dataset.add_files(
        File.from_remote(
            remote_path="source/file1.h5", size=512, creation_time=dataset.creation_time
        )
    )

    client = FakeClient()
    finalized = client.upload_new_dataset_now(dataset)
    pid = cast(PID, finalized.pid)
    expected = client.get_dataset(pid, attachments=True).replace(
        # The backend may update the dataset after upload
        _read_only={
            "updated_at": finalized.updated_at,
            "updated_by": finalized.updated_by,
        }
    )
    assert finalized == expected


def test_upload_with_both_remote_and_local_files(client, dataset_with_files):
    original_file_names = set(
        dataset_with_files.source_folder / file.remote_path
        for file in dataset_with_files.files
    )
    dataset_with_files.add_files(
        File.from_remote(
            remote_path="file1.h5", size=6123, creation_time="2019-09-09T19:29:39Z"
        )
    )

    finalized = client.upload_new_dataset_now(dataset_with_files)
    expected = client.get_dataset(finalized.pid, attachments=True).replace(
        # The backend may update the dataset after upload
        _read_only={
            "updated_at": finalized.updated_at,
            "updated_by": finalized.updated_by,
        }
    )
    assert finalized == expected
    # Does not include the remote file "file1.h5"
    assert get_file_transfer(client).files.keys() == original_file_names


def test_upload_with_file_with_both_remote_and_local_path(client, dataset_with_files):
    file = File.from_remote(
        remote_path="file1.h5", size=6123, creation_time="2019-09-09T19:29:39Z"
    )
    file = file.downloaded(local_path="/local/file1.h5")
    dataset_with_files.add_files(file)

    # Client w/o file transfer to ensure that the client never attempts to upload.
    client = FakeClient()
    with pytest.raises(ValueError, match="path"):
        client.upload_new_dataset_now(dataset_with_files)


def test_upload_creates_dataset_and_datablock(client, dataset_with_files):
    finalized = client.upload_new_dataset_now(dataset_with_files)
    assert client.datasets[finalized.pid].createdAt == finalized.created_at
    assert client.datasets[finalized.pid].datasetName == finalized.name
    assert client.datasets[finalized.pid].owner == finalized.owner
    assert client.datasets[finalized.pid].size == finalized.size

    assert client.orig_datablocks[finalized.pid][0].createdBy == finalized.created_by
    assert client.orig_datablocks[finalized.pid][0].datasetId == finalized.pid
    assert client.orig_datablocks[finalized.pid][0].size == finalized.size


def test_upload_creates_attachments(client, dataset, attachments):
    dataset.attachments = attachments
    finalized = client.upload_new_dataset_now(dataset)

    uploaded = client.attachments[finalized.pid]
    assert len(uploaded) == len(attachments)
    assert uploaded[0].caption == attachments[0].caption
    assert uploaded[0].ownerGroup == attachments[0].owner_group
    assert uploaded[0].thumbnail == attachments[0].thumbnail
    assert uploaded[1].caption == attachments[1].caption
    assert uploaded[1].ownerGroup == attachments[1].owner_group
    assert uploaded[1].thumbnail == attachments[1].thumbnail


def test_upload_uploads_files_to_source_folder(client, dataset_with_files):
    finalized = client.upload_new_dataset_now(dataset_with_files)
    source_folder = client.file_transfer.source_folder_for(finalized)

    assert (
        get_file_transfer(client).files[source_folder / "file.nxs"]
        == b"contents of file.nxs"
    )
    assert (
        get_file_transfer(client).files[source_folder / "the_log_file.log"]
        == b"this is a log file"
    )


def test_upload_does_not_create_dataset_if_file_upload_fails(dataset_with_files, fs):
    class RaisingUpload(FakeFileTransfer):
        source_dir = "/"

        def upload_files(self, *files):
            raise RuntimeError("Fake upload failure")

        @contextmanager
        def connect_for_upload(self, pid):
            yield self

    client = FakeClient(file_transfer=RaisingUpload(fs=fs))

    with pytest.raises(RuntimeError, match="Fake upload failure"):
        client.upload_new_dataset_now(dataset_with_files)

    assert not client.datasets
    assert not client.orig_datablocks
    assert not client.attachments


def test_upload_cleans_up_files_if_dataset_ingestion_fails(dataset_with_files, fs):
    client = FakeClient(
        disable={"create_dataset_model": ScicatCommError("Ingestion failed")},
        file_transfer=FakeFileTransfer(fs=fs),
    )
    with pytest.raises(ScicatCommError):
        client.upload_new_dataset_now(dataset_with_files)

    assert not get_file_transfer(client).files


def test_upload_does_not_create_dataset_if_validation_fails(dataset_with_files, fs):
    client = FakeClient(
        disable={"validate_dataset_model": ValueError("Validation failed")},
        file_transfer=FakeFileTransfer(fs=fs),
    )
    with pytest.raises(ValueError):
        client.upload_new_dataset_now(dataset_with_files)

    assert not client.datasets
    assert not client.orig_datablocks
    assert not client.attachments
    assert not get_file_transfer(client).files


def test_failed_datablock_upload_does_not_revert(dataset_with_files, fs):
    client = FakeClient(
        disable={"create_orig_datablock": ScicatCommError("Ingestion failed")},
        file_transfer=FakeFileTransfer(fs=fs),
    )
    with pytest.raises(RuntimeError):
        client.upload_new_dataset_now(dataset_with_files)

    uploaded_dset = next(iter(client.datasets.values()))
    assert uploaded_dset.owner == "PonderStibbons"
    assert uploaded_dset.datasetName == "Data A38"
    assert uploaded_dset.usedSoftware == ["EasyScience"]

    assert get_file_transfer(client).files
    assert not get_file_transfer(client).reverted


def test_upload_does_not_create_attachments_if_dataset_ingestion_fails(
    attachments, dataset
):
    dataset.attachments = attachments
    client = FakeClient(
        disable={"create_dataset_model": ScicatCommError("Ingestion failed")},
        file_transfer=FakeFileTransfer(),
    )
    with pytest.raises(ScicatCommError):
        client.upload_new_dataset_now(dataset)

    assert not client.attachments


def test_upload_does_not_create_attachments_if_datablock_ingestion_fails(
    attachments, dataset_with_files, fs
):
    dataset_with_files.attachments = attachments
    client = FakeClient(
        disable={"create_orig_datablock": ScicatCommError("Ingestion failed")},
        file_transfer=FakeFileTransfer(fs=fs),
    )
    with pytest.raises(RuntimeError):
        client.upload_new_dataset_now(dataset_with_files)

    assert not client.attachments


def test_failed_attachment_upload_does_not_revert(attachments, dataset_with_files, fs):
    dataset_with_files.attachments = attachments
    client = FakeClient(
        disable={"create_attachment_for_dataset": ScicatCommError("Ingestion failed")},
        file_transfer=FakeFileTransfer(fs=fs),
    )
    with pytest.raises(RuntimeError):
        client.upload_new_dataset_now(dataset_with_files)

    uploaded_dset = next(iter(client.datasets.values()))
    assert uploaded_dset.owner == "PonderStibbons"
    assert uploaded_dset.datasetName == "Data A38"
    assert uploaded_dset.usedSoftware == ["EasyScience"]

    (uploaded_dblock,) = next(iter(client.orig_datablocks.values()))
    assert uploaded_dblock.datasetId == uploaded_dset.pid

    assert get_file_transfer(client).files
    assert not get_file_transfer(client).reverted



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

"""Wrappers for (Orig)Datablocks."""

from __future__ import annotations

import dataclasses
from datetime import datetime
from typing import TYPE_CHECKING, Iterable, Iterator, List, Optional

from .file import File
from .model import DownloadOrigDatablock, UploadOrigDatablock
from .pid import PID

if TYPE_CHECKING:
    from .dataset import Dataset

# TODO Datablock


@dataclasses.dataclass
class OrigDatablock:
    """Dataclass for an orig datablock.

    Instances of this class are mutable as opposed to
    :class:`scitacean.model.OrigDatablock`.
    They are used for building datasets and get converted to/from pydantic
    models for communication with a server.
    """

    _files: List[File] = dataclasses.field(init=False)
    checksum_algorithm: Optional[str] = None
    access_groups: Optional[List[str]] = None
    instrument_group: Optional[str] = None
    owner_group: Optional[str] = None
    init_files: dataclasses.InitVar[Optional[Iterable[File]]] = None
    _created_at: Optional[datetime] = None
    _created_by: Optional[str] = None
    _dataset_id: Optional[PID] = None
    _id: Optional[str] = None
    _updated_at: Optional[datetime] = None
    _updated_by: Optional[str] = None

    def __post_init__(self, init_files: Optional[Iterable[File]]) -> None:
        self._files = list(init_files) if init_files is not None else []

    @classmethod
    def from_download_model(
        cls,
        orig_datablock_model: DownloadOrigDatablock,
    ) -> OrigDatablock:
        """Construct a new OrigDatablock from pydantic models.

        Parameters
        ----------
        orig_datablock_model:
            Model of the orig datablock to construct.

        Returns
        -------
        :
            A new instance.
        """
        dblock = orig_datablock_model
        return OrigDatablock(
            checksum_algorithm=dblock.chkAlg,
            owner_group=dblock.ownerGroup,
            access_groups=dblock.accessGroups,
            instrument_group=dblock.instrumentGroup,
            _created_at=dblock.createdAt,
            _created_by=dblock.createdBy,
            _dataset_id=orig_datablock_model.datasetId,
            _id=orig_datablock_model.id,
            _updated_at=dblock.updatedAt,
            _updated_by=dblock.updatedBy,
            init_files=[
                File.from_download_model(file, checksum_algorithm=dblock.chkAlg)
                for file in dblock.dataFileList or ()
            ],
        )

    @property
    def files(self) -> Iterator[File]:
        """Iterator over all files."""
        return iter(self._files)

    @property
    def size(self) -> int:
        """Total size of all files."""
        return sum(file.size for file in self.files)

    @property
    def created_at(self) -> Optional[datetime]:
        """Creation time of this orig datablock."""
        return self._created_at

    @property
    def created_by(self) -> Optional[str]:
        """User who created this orig datablock."""
        return self._created_by

    @property
    def updated_at(self) -> Optional[datetime]:
        """Last update time of this orig datablock."""
        return self._updated_at

    @property
    def updated_by(self) -> Optional[str]:
        """User who last updated this datablock."""
        return self._updated_by

    @property
    def dataset_id(self) -> Optional[PID]:
        """PID of the dataset this datablock belongs to."""
        return self._dataset_id

    @property
    def datablock_id(self) -> Optional[str]:
        """ID of this datablock."""
        return self._id

    def add_files(self, *files: File) -> None:
        """Append files to the datablock.

        Parameters
        ----------
        files:
            File objects to add.
        """
        self._files.extend(
            dataclasses.replace(f, checksum_algorithm=self.checksum_algorithm)
            for f in files
        )

    def make_upload_model(self, dataset: Dataset) -> UploadOrigDatablock:
        """Build a new pydantic model to upload this datablock.

        Parameters
        ----------
        dataset:
            The dataset that this orig datablock belongs to.

        Returns
        -------
        :
            A new model for this orig datablock.
        """
        owner_group = self.owner_group or dataset.owner_group
        return UploadOrigDatablock(
            chkAlg=self.checksum_algorithm,
            size=self.size,
            dataFileList=[file.make_model(for_archive=False) for file in self.files],
            datasetId=dataset.pid,  # type: ignore[arg-type]
            ownerGroup=owner_group,
            accessGroups=self.access_groups or dataset.access_groups,
            instrumentGroup=self.instrument_group or dataset.instrument_group,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

import pickle

import pytest

from scitacean import PID, Client
from scitacean.testing.client import FakeClient
from scitacean.util.credentials import SecretStr


def test_from_token_fake():
    # This should not call the API
    client = FakeClient.from_token(url="some.url/api/v3", token="a-token")  # noqa: S106
    assert isinstance(client, FakeClient)


def test_from_credentials_fake():
    # This should not call the API
    client = FakeClient.from_credentials(
        url="some.url/api/v3",
        username="someone",
        password="the-fake-does-not-care",  # noqa: S106
    )
    assert isinstance(
        client,
        FakeClient,
    )


def test_from_credentials_real(scicat_access, scicat_backend):
    if not scicat_backend:
        pytest.skip("No backend")
    Client.from_credentials(url=scicat_access.url, **scicat_access.user.credentials)


def test_cannot_pickle_client_credentials_manual_token_str():
    client = Client.from_token(url="/", token="the-token")  # noqa: S106
    with pytest.raises(TypeError):
        pickle.dumps(client)


def test_cannot_pickle_client_credentials_manual_token_secret_str():
    client = Client.from_token(url="/", token=SecretStr("the-token"))
    with pytest.raises(TypeError):
        pickle.dumps(client)


def test_cannot_pickle_client_credentials_login(scicat_access, require_scicat_backend):
    client = Client.from_credentials(
        url=scicat_access.url, **scicat_access.user.credentials
    )
    with pytest.raises(TypeError):
        pickle.dumps(client)


def test_connection_error_does_not_contain_token():
    client = Client.from_token(
        url="https://not-actually-a_server",
        token="the token/which_must-be.kept secret",  # noqa: S106
    )
    try:
        client.get_dataset("does not exist")
        assert False, "There must be an exception"  # noqa: B011
    except Exception as exc:
        assert "the token/which_must-be.kept secret" not in str(exc)
        for arg in exc.args:
            assert "the token/which_must-be.kept secret" not in str(arg)


def test_fake_can_disable_functions():
    client = FakeClient(
        disable={
            "get_dataset_model": RuntimeError("custom failure"),
            "get_orig_datablocks": IndexError("custom index error"),
        }
    )
    with pytest.raises(RuntimeError, match="custom failure"):
        client.scicat.get_dataset_model(PID(pid="some-pid"))
    with pytest.raises(IndexError, match="custom index error"):
        client.scicat.get_orig_datablocks(PID(pid="some-pid"))
